# Form implementation generated from reading ui file 'searchbar.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_SearchBar(object):
    def setupUi(self, SearchBar):
        SearchBar.setObjectName("SearchBar")
        SearchBar.resize(329, 26)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SearchBar.sizePolicy().hasHeightForWidth())
        SearchBar.setSizePolicy(sizePolicy)
        SearchBar.setAutoFillBackground(True)
        self.horizontalLayout = QHBoxLayout(SearchBar)
        self.horizontalLayout.setContentsMargins(2, 2, 2, 2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.providerChooser = QToolButton(parent=SearchBar)
        self.providerChooser.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.providerChooser.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.providerChooser.setObjectName("providerChooser")
        self.horizontalLayout.addWidget(self.providerChooser)
        self.lineEdit = QLineEdit(parent=SearchBar)
        self.lineEdit.setClearButtonEnabled(True)
        self.lineEdit.setObjectName("lineEdit")
        self.horizontalLayout.addWidget(self.lineEdit)
        self.filterCheckBox = QCheckBox(parent=SearchBar)
        self.filterCheckBox.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.filterCheckBox.setObjectName("filterCheckBox")
        self.horizontalLayout.addWidget(self.filterCheckBox)
        self.forwardButton = QToolButton(parent=SearchBar)
        self.forwardButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.forwardButton.setText("↓")
        self.forwardButton.setObjectName("forwardButton")
        self.horizontalLayout.addWidget(self.forwardButton)
        self.backwardButton = QToolButton(parent=SearchBar)
        self.backwardButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.backwardButton.setText("↑")
        self.backwardButton.setObjectName("backwardButton")
        self.horizontalLayout.addWidget(self.backwardButton)
        self.closeButton = QToolButton(parent=SearchBar)
        self.closeButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.closeButton.setObjectName("closeButton")
        self.horizontalLayout.addWidget(self.closeButton)

        self.retranslateUi(SearchBar)

    def retranslateUi(self, SearchBar):
        self.providerChooser.setToolTip(_p("SearchBar", "Search scope"))
        self.filterCheckBox.setToolTip(_p("SearchBar", "Show only matching rows"))
        self.filterCheckBox.setText(_p("SearchBar", "Filter"))
        self.forwardButton.setToolTip(_p("SearchBar", "Next Occurrence"))
        self.backwardButton.setToolTip(_p("SearchBar", "Previous Occurrence"))
        self.closeButton.setToolTip(_p("SearchBar", "Close Search Bar"))
        self.closeButton.setText(_p("SearchBar", "Done"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Translated tables of enums, exception names, preference keys.
"""

from __future__ import annotations

import os
import re
import textwrap
from enum import Enum
from collections.abc import Callable

from gitfourchette.appconsts import *
from gitfourchette.localization import *

_userCommandsGuideUrl = "https://gitfourchette.org/guide/commands"


def retranslate(newLanguageCode: str):
    """Flush cached translations."""
    # Bump table cookie value, forcing _getTable to regenerate all tables.
    global _tableCookieValue
    _tableCookieValue = newLanguageCode


def enum(enumValue: Enum) -> str:
    """Translate an enum value."""
    table = _getTable(_enumTable)
    try:
        return table[type(enumValue)][enumValue]
    except KeyError:
        return enumValue.name


def exceptionName(exc: BaseException) -> str:
    """Translate an exception name."""
    table = _getTable(_exceptionNameTable)
    name = type(exc).__name__
    return table.get(name, name)


def prefKey(key: str, default: str | None = None) -> str:
    """Translate a PrefsFile field."""
    table = _getTable(_prefKeyTable)
    try:
        return table[key]
    except KeyError:
        return str(key) if default is None else default


def prefKeyNoDefault(key: str) -> str:
    """Translate a PrefsFile field (return empty string if no translation
    available)."""
    return prefKey(key, "")


# -------------------------------------------------------------------------
# Table caching
# -------------------------------------------------------------------------

_tableAttr = "_cachedLocalizationTable"
_tableCookieKey = "_TrTablesLanguageCheck"
_tableCookieValue = ""


def _getTable[T: dict](initializer: Callable[[], T]) -> T:
    # See if we've stashed a table in this initializer's attributes previously
    table: T | None = getattr(initializer, _tableAttr, None)

    # No table, or table cookie is stale?
    if table is None or table[_tableCookieKey] != _tableCookieValue:
        # Generate the table
        table = initializer()

        # Set fresh cookie
        table[_tableCookieKey] = _tableCookieValue

        # Stash the table inside the initializer
        setattr(initializer, _tableAttr, table)

    assert table is not None
    return table


# -------------------------------------------------------------------------
# Table initializers
# -------------------------------------------------------------------------

def _exceptionNameTable() -> dict[str, str]:
    return {
        "ConnectionRefusedError": _("Connection refused"),
        "FileNotFoundError": _("File not found"),
        "GitError": _("Git error"),
        "InterruptedError": _("Operation interrupted"),
        "NotImplementedError": _("Unsupported feature"),
        "PermissionError": _("Permission denied"),
    }


def _enumTable() -> dict[type[Enum], dict[Enum, str]]:
    from gitfourchette.gitdriver import GitConflictSides, GitStatus
    from gitfourchette.nav import NavContext
    from gitfourchette.porcelain import FileMode, NameValidationError, RepositoryState
    from gitfourchette.toolbox import toLengthVariants
    from gitfourchette.sidebar.sidebarmodel import SidebarItem
    from gitfourchette.settings import (
        WhitespaceMode,
        FileListClick,
        GraphRefBoxWidth,
        GraphRowHeight,
        QtApiNames,
        RefSort,
        TabBarClick,
    )
    from gitfourchette.themes import ThemeAccent, ThemeName
    from gitfourchette.toolbox import PatchPurpose, PathDisplayStyle, AuthorDisplayStyle
    from gitfourchette.repomodel import GpgStatus

    NVERule = NameValidationError.Rule

    table: dict[type[Enum], dict[Enum, str]] = {}

    table[AuthorDisplayStyle] = {
        AuthorDisplayStyle.FullName: _("Full name"),
        AuthorDisplayStyle.FirstName: _("First name"),
        AuthorDisplayStyle.LastName: _("Last name"),
        AuthorDisplayStyle.Initials: _("Initials"),
        AuthorDisplayStyle.FullEmail: _("Full email"),
        AuthorDisplayStyle.EmailUserName: _("Abbreviated email"),
    }

    table[FileListClick] = {
        FileListClick.Nothing: _("Do nothing"),
        FileListClick.Stage: _("Stage/unstage file"),
        FileListClick.Blame: _("Blame file"),
        FileListClick.Edit: _("Open file in external editor"),
        FileListClick.DiffTool: _("Open file in external diff tool"),
        FileListClick.Folder: _("Show file in folder"),
    }

    table[FileMode] = {
        FileMode.UNREADABLE: _p("FileMode", "deleted"),
        FileMode.BLOB: _p("FileMode", "regular file"),
        FileMode.BLOB_EXECUTABLE: _p("FileMode", "executable file"),
        FileMode.LINK: _p("FileMode", "symbolic link"),
        FileMode.TREE: _p("FileMode", "subtree"),
        FileMode.COMMIT: _p("FileMode", "subtree commit"),
    }

    table[GitConflictSides] = {
        GitConflictSides.BothDeleted: _p("ConflictSides", "deleted by both sides"),
        GitConflictSides.AddedByUs: _p("ConflictSides", "added by us"),
        GitConflictSides.DeletedByThem: _p("ConflictSides", "deleted by them"),
        GitConflictSides.AddedByThem: _p("ConflictSides", "added by them"),
        GitConflictSides.DeletedByUs: _p("ConflictSides", "deleted by us"),
        GitConflictSides.BothAdded: _p("ConflictSides", "added by both sides"),
        GitConflictSides.BothModified: _p("ConflictSides", "modified by both sides"),
    }

    table[GitStatus] = {
        GitStatus.Added: _p("FileStatus", "added"),
        GitStatus.Copied: _p("FileStatus", "copied"),
        GitStatus.Deleted: _p("FileStatus", "deleted"),
        GitStatus.Ignored: _p("FileStatus", "ignored"),
        GitStatus.Modified: _p("FileStatus", "modified"),
        GitStatus.Renamed: _p("FileStatus", "renamed"),
        GitStatus.TypeChanged: _p("FileStatus", "file type changed"),
        GitStatus.Unmerged: _p("FileStatus", "merge conflict"),
        GitStatus.Unknown: _p("FileStatus", "unreadable"),
        GitStatus.Untracked: _p("FileStatus", "untracked"),
    }

    table[GpgStatus] = {
        GpgStatus.Unsigned: _("Not signed"),
        GpgStatus.Pending: _("Signature not verified yet"),
        GpgStatus.CantCheck: _("Unable to verify signature"),
        GpgStatus.MissingKey: _("Can’t verify signature; Key not in your keyring"),
        GpgStatus.GoodTrusted: _("Good signature; Key trusted"),
        GpgStatus.GoodUntrusted: _("Good signature; Key not fully trusted"),
        GpgStatus.ExpiredSig: _("Good signature; Signature expired"),
        GpgStatus.ExpiredKey: _("Good signature; Key expired"),
        GpgStatus.RevokedKey: _("Good signature; Key revoked"),
        GpgStatus.Bad: _("Bad signature"),
        GpgStatus.ProcessError: _("Failed to start verification process"),
    }

    table[GraphRefBoxWidth] = {
        GraphRefBoxWidth.IconsOnly: _("Icons Only"),
        GraphRefBoxWidth.Standard: _("Truncate long ref names"),
        GraphRefBoxWidth.Wide: _("Show full ref names"),
    }

    table[GraphRowHeight] = {
        GraphRowHeight.Cramped: _p("row spacing", "Cramped"),
        GraphRowHeight.Tight: _p("row spacing", "Tight"),
        GraphRowHeight.Relaxed: _p("row spacing", "Relaxed"),
        GraphRowHeight.Roomy: _p("row spacing", "Roomy"),
        GraphRowHeight.Spacious: _p("row spacing", "Spacious"),
    }

    table[NavContext] = {
        NavContext.EMPTY: _p("NavContext", "Empty"),
        NavContext.UNSTAGED: _p("NavContext", "Unstaged"),
        NavContext.STAGED: _p("NavContext", "Staged"),
        NavContext.COMMITTED: _p("NavContext", "Committed"),
    }

    table[NVERule] = {
        NVERule.ILLEGAL_NAME: _("Illegal name."),
        NVERule.ILLEGAL_SUFFIX: _("Illegal suffix."),
        NVERule.ILLEGAL_PREFIX: _("Illegal prefix."),
        NVERule.CONTAINS_ILLEGAL_SEQ: _("Contains illegal character sequence."),
        NVERule.CONTAINS_ILLEGAL_CHAR: _("Contains illegal character."),
        NVERule.CANNOT_BE_EMPTY: _("Cannot be empty."),
        NVERule.NOT_WINDOWS_FRIENDLY: _("This name is discouraged for compatibility with Windows."),
        NVERule.NAME_TAKEN_BY_REF: _("This name is already taken."),
        NVERule.NAME_TAKEN_BY_FOLDER: _("This name is already taken by a folder."),
        NVERule.NOT_A_FOLDER: _("The folder in this path clashes with an existing ref that isn’t a folder."),
    }

    table[PatchPurpose] = {
        PatchPurpose.Stage: _p("PatchPurpose", "Stage"),
        PatchPurpose.Unstage: _p("PatchPurpose", "Unstage"),
        PatchPurpose.Discard: _p("PatchPurpose", "Discard"),
        PatchPurpose.Lines | PatchPurpose.Stage: _p("PatchPurpose", "Stage lines"),
        PatchPurpose.Lines | PatchPurpose.Unstage: _p("PatchPurpose", "Unstage lines"),
        PatchPurpose.Lines | PatchPurpose.Discard: _p("PatchPurpose", "Discard lines"),
        PatchPurpose.Hunk | PatchPurpose.Stage: _p("PatchPurpose", "Stage hunk"),
        PatchPurpose.Hunk | PatchPurpose.Unstage: _p("PatchPurpose", "Unstage hunk"),
        PatchPurpose.Hunk | PatchPurpose.Discard: _p("PatchPurpose", "Discard hunk"),
        PatchPurpose.File | PatchPurpose.Stage: _p("PatchPurpose", "Stage file"),
        PatchPurpose.File | PatchPurpose.Unstage: _p("PatchPurpose", "Unstage file"),
        PatchPurpose.File | PatchPurpose.Discard: _p("PatchPurpose", "Discard file"),
    }

    table[PathDisplayStyle] = {
        PathDisplayStyle.FullPaths: _("Full paths"),
        PathDisplayStyle.AbbreviateDirs: _("Abbreviate directories"),
        PathDisplayStyle.FileNameOnly: _("Filename only"),
        PathDisplayStyle.FileNameFirst: _("Filename first"),
    }

    table[QtApiNames] = {
        QtApiNames.Automatic: _p("Qt binding", "Automatic (recommended)"),
        QtApiNames.PySide6: "PySide6",
        QtApiNames.PyQt6: "PyQt6",
        QtApiNames.PyQt5: "PyQt5"
    }

    table[RefSort] = {
        RefSort.TimeDesc: _p("sort refs by date of latest commit, descending", "Date, Newest First"),
        RefSort.TimeAsc: _p("sort refs by date of latest commit, ascending", "Date, Oldest First"),
        RefSort.AlphaAsc: _p("sort refs alphabetically, ascending", "Name, A-Z"),
        RefSort.AlphaDesc: _p("sort refs alphabetically, descending", "Name, Z-A"),
        RefSort.UseGlobalPref: "",
    }

    table[RepositoryState] = {
        RepositoryState.NONE: _p("RepositoryState", "None"),
        RepositoryState.MERGE: _p("RepositoryState", "Merging"),
        RepositoryState.REVERT: _p("RepositoryState", "Reverting"),
        RepositoryState.REVERT_SEQUENCE: _p("RepositoryState", "Reverting (sequence)"),
        RepositoryState.CHERRYPICK: _p("RepositoryState", "Cherry-picking"),
        RepositoryState.CHERRYPICK_SEQUENCE: _p("RepositoryState", "Cherry-picking (sequence)"),
        RepositoryState.BISECT: _p("RepositoryState", "Bisecting"),
        RepositoryState.REBASE: _p("RepositoryState", "Rebasing"),
        RepositoryState.REBASE_INTERACTIVE: _p("RepositoryState", "Rebasing (interactive)"),
        RepositoryState.REBASE_MERGE: _p("RepositoryState", "Rebasing (merging)"),
        RepositoryState.APPLY_MAILBOX: "Apply Mailbox",  # intentionally untranslated
        RepositoryState.APPLY_MAILBOX_OR_REBASE: "Apply Mailbox or Rebase",  # intentionally untranslated
    }

    table[SidebarItem] = {
        SidebarItem.UncommittedChanges: toLengthVariants(_p("SidebarModel", "Working Directory|Workdir")),
        SidebarItem.LocalBranchesHeader: toLengthVariants(_p("SidebarModel", "Local Branches|Branches")),
        SidebarItem.StashesHeader: _p("SidebarModel", "Stashes"),
        SidebarItem.RemotesHeader: _p("SidebarModel", "Remotes"),
        SidebarItem.TagsHeader: _p("SidebarModel", "Tags"),
        SidebarItem.SubmodulesHeader: _p("SidebarModel", "Submodules"),
        SidebarItem.LocalBranch: _p("SidebarModel", "Local branch"),
        SidebarItem.DetachedHead: _p("SidebarModel", "Detached HEAD"),
        SidebarItem.UnbornHead: _p("SidebarModel", "Unborn HEAD"),
        SidebarItem.RemoteBranch: _p("SidebarModel", "Remote branch"),
        SidebarItem.Stash: _p("SidebarModel", "Stash"),
        SidebarItem.Remote: _p("SidebarModel", "Remote"),
        SidebarItem.Tag: _p("SidebarModel", "Tag"),
        SidebarItem.Submodule: _p("SidebarModel", "Submodules"),
        SidebarItem.Spacer: "---",
    }

    table[TabBarClick] = {
        TabBarClick.Nothing: _("Do nothing"),
        TabBarClick.Folder: _("Open repo folder"),
        TabBarClick.Terminal: _("Open repo in terminal"),
        TabBarClick.Close: _("Close tab"),
    }

    table[ThemeAccent] = {
        ThemeAccent.Blue: _("Blue"),
        ThemeAccent.Cyan: _("Cyan"),
        ThemeAccent.Green: _("Green"),
        ThemeAccent.Yellow: _("Yellow"),
        ThemeAccent.Orange: _("Orange"),
        ThemeAccent.Red: _("Red"),
        ThemeAccent.Pink: _("Pink"),
        ThemeAccent.Gray: _("Gray"),
        ThemeAccent.Indigo: _("Indigo"),
        ThemeAccent.Purple: _("Purple"),
    }

    table[ThemeName] = {
        ThemeName.BuiltIn: APP_DISPLAY_NAME,
    }

    table[WhitespaceMode] = {
        WhitespaceMode.Strict: _("Do not ignore whitespace changes"),
        WhitespaceMode.IgnoreAll: _("Ignore all whitespace changes"),
        WhitespaceMode.IgnoreChange: _("Ignore length changes in existing whitespace"),
        WhitespaceMode.IgnoreCrAtEol: _("Ignore line ending changes only ({0})", "LF ↔ CRLF"),
    }

    return table


def _prefKeyTable() -> dict[str, str]:
    from gitfourchette.toolbox.textutils import paragraphs, tquo, escape
    from gitfourchette.exttools.usercommand import UserCommand

    sshAuthSock = os.environ.get("SSH_AUTH_SOCK", "")
    if sshAuthSock:
        sshAuthSockHelp = paragraphs(
            _("Note: Per {k}, your system is providing an ssh-agent ({v}). "
              "It’s recommended to use this one."),
            _("If your system’s agent isn’t saving any passphrases, "
              "make sure you’ve enabled {c} in your SSH configuration."))
    else:
        sshAuthSockHelp = _("Note: Per {k}, no ssh-agent seems to be running on your system.")
    sshAuthSockHelp = "<blockquote>" + sshAuthSockHelp.format(
        k="SSH_AUTH_SOCK", v=escape(sshAuthSock), c="AddKeysToAgent")

    return {
        "general": _p("Prefs", "General"),
        "diff": _p("Prefs", "Code"),
        "imageDiff": _p("Prefs", "Images"),
        "tabs": _p("Prefs", "Tabs"),
        "graph": _p("Prefs", "Commit History"),
        "trash": _p("Prefs", "Trash"),
        "git": _p("Prefs", "Git Integration"),
        "external": _p("Prefs", "External Tools"),
        "advanced": _p("Prefs", "Advanced"),
        "userCommands": _p("Prefs", "Custom Commands"),

        "language": _("Language"),
        "qtStyle": _("Qt style"),
        "shortHashChars": _("Shorten hashes to # characters"),
        "shortTimeFormat": _("Date/time format"),
        "shortTimeFormat_help": _timeFormatGuide(),
        "pathDisplayStyle": _("Path display style"),
        "authorDisplayStyle": _("Author display style"),
        "maxRecentRepos": _("Remember up to # recent repositories"),
        "showStatusBar": _("Show status bar"),
        "showToolBar": _("Show toolbar"),
        "showMenuBar": _("Show menu bar"),
        "showMenuBar_help": _("When the menu bar is hidden, press the Alt key to show it again."),
        "resetDontShowAgain": _("Restore all “don’t show this again” messages"),
        "pygmentsPlugins": _("Allow third-party Pygments plugins"),
        "pygmentsPlugins_help": "<p>" + _("Let {app} load third-party Pygments plugins installed on your system. "
                                          "These plugins extend syntax highlighting with new languages "
                                          "and color schemes. <b>May incur significant slowdowns.</b>"),
        "refSort": _("Sort branches && tags by"),
        "refSort_help": paragraphs(
            _("The default sorting mode for local branches, remote branches, and tags in the sidebar."),
            _("You can fine-tune this setting in each repo by right-clicking Branches, Remotes, or Tags "
              "in the sidebar. (Note that changing the default setting here will clear per-repo tweaks.)")),

        "font": _("Font"),
        "tabSpaces": _("One tab is # spaces"),
        "contextLines": _("Show up to # context lines"),
        "contextLines_help": _("Amount of unmodified lines to show around red or green lines in a diff."),
        "largeFileThresholdKB": _("Load diffs up to # KB"),
        "imageFileThresholdKB": _("Load images up to # KB"),
        "syntaxHighlighting": _("Syntax highlighting"),
        "wordWrap": _("Word wrap"),
        "showStrayCRs": _("Show alien line endings (CRLF)"),
        "showWhitespace": _("Show whitespace symbols"),
        "whitespaceMode": _("Whitespace diffs"),
        "whitespaceMode_help": paragraphs(
            _("How whitespace changes are shown in diffs."),
            _("This setting only affects how diffs are displayed. "
              "Operations that handle patches (export, apply, revert, etc.) still honor all whitespace changes.")),
        "colorblind": _("“-/+” colors"),
        "colorblind_help": _("Background colors for deleted (-) and added (+) lines."),
        "renderSvg": _("SVG files"),
        "renderSvg_false": _("Display as text"),
        "renderSvg_true": _("Display as images"),

        "tabCloseButton": _("Show tab close button"),
        "expandingTabs": _("Expand tabs to available width"),
        "autoHideTabs": _("Auto-hide tabs if only one repo is open"),

        "chronologicalOrder": _("Sort commits"),
        "chronologicalOrder_true": _("Chronologically"),
        "chronologicalOrder_false": _("Topologically"),
        "chronologicalOrder_help": paragraphs(
            _("<b>Chronological mode</b> lets you stay on top of the latest activity in the repository. "
              "The most recent commits always show up at the top of the graph. "
              "However, the graph can get messy when multiple branches receive commits in the same timeframe."),
            _("<b>Topological mode</b> makes the graph easier to read. It attempts to present sequences of "
              "commits within a branch in a linear fashion. Since this is not a strictly chronological "
              "mode, you may have to do more scrolling to see the latest changes in various branches."),
        ),
        "graphRowHeight": _("Row spacing"),
        "flattenLanes": _("Avoid gaps between branches in the graph"),
        "authorDiffAsterisk": _("Mark author/committer signature differences"),
        "authorDiffAsterisk_help": paragraphs(
            _("The commit history displays information about a commit’s <b>author</b>—"
              "their name and the date at which they made the commit. But in some cases, a commit "
              "might have been revised by someone else than the original author—"
              "this person is called the <b>committer</b>."),
            _("If you tick this option, an asterisk (*) will appear after the author’s name "
              "and/or date if they differ from the committer’s for any given commit."),
            _("Note that you can always hover over the author’s name or date to obtain "
              "detailed information about the author and the committer."),
        ),
        "maxCommits": _("Load up to # commits in the history"),
        "maxCommits_help": _("Set to 0 to always load the full commit history."),
        "alternatingRowColors": _("Draw rows using alternating background colors"),
        "refBoxMaxWidth": _("Ref indicators"),
        "refBoxMaxWidth_help": _("You can always hover over an indicator to display the full name of the ref."),
        "verifyGpgOnTheFly": _("Verify signed commits on the fly"),
        "verifyGpgOnTheFly_help": _(
            "As commits scroll into view, call {0} automatically to verify their signatures. "
            "The verification status is materialized by a seal icon next to the author’s name:", tquo("git verify-commit")
        ) + _gpgStatusReferenceTable() + "<br>" + _("(No seal = Commit isn’t signed)"),

        "maxTrashFiles": _("The trash keeps up to # discarded patches"),
        "maxTrashFileKB": _("Patches bigger than # KB won’t be salvaged"),
        "trash_HEADER": _(
            "When you discard changes from the working directory, {app} keeps a temporary copy in a hidden "
            "“trash” folder. This gives you a last resort to rescue changes that you have discarded by mistake. "
            "You can look around this trash folder via <i>“Help &rarr; Open Trash”</i>."),

        "verbosity": _("Logging verbosity"),
        "autoRefresh": _("Auto-refresh when app regains focus"),
        "autoRefresh_help": paragraphs(
            _("When you return to {app} from another application, it automatically "
              "scans for changes in the working directory and local branches. "
              "This keeps the interface in sync with the state of your repo on disk."),
            _("If you turn this off, you will need to hit {key} to "
              "perform this refresh manually.", key="F5"),
            "<b>" + _("We strongly recommend to keep this setting enabled.") + "</b>"),
        "autoFetchMinutes": _("Auto-fetch remotes every # minutes"),
        "animations": _("Animation effects in sidebar"),
        "smoothScroll": _("Smooth scrolling (where applicable)"),
        "forceQtApi": _("Preferred Qt binding"),
        "forceQtApi_help": paragraphs(
            _("After restarting, {app} will use this Qt binding if available."),
            _("You can also pass the name of a Qt binding via the “QT_API” environment variable."),
        ),
        "condensedFonts": _("Use condensed fonts"),
        "condensedFonts_help": "<p>" + _(
            "When a branch name or author name is too long to fit in its allotted space, "
            "condense the font before truncating the text."),

        "externalEditor": _("Text editor"),
        "externalDiff": _("Diff tool"),
        "externalDiff_help":
            "<p style='white-space: pre'>" + _("Argument placeholders:") + "\n" + _tokenReferenceTable({
                "$L": _("Old / Left"),
                "$R": _("New / Right"),
            }),
        "externalMerge": _("Merge tool"),
        "externalMerge_help":
            "<p style='white-space: pre'>" + _("Argument placeholders:") + "\n" + _tokenReferenceTable({
                "$B": _("Ancestor / Base / Center"),
                "$L": _("Ours / Local / Left"),
                "$R": _("Theirs / Remote / Right"),
                "$M": _("Merged / Output / Result"),
            }),
        "terminal": _("Terminal"),
        "terminal_help": paragraphs(
            _("Argument placeholders:"),
            _tokenReferenceTable({"$COMMAND": _("Command to execute after launching the terminal")}),
            _("The {0} placeholder is mandatory. It is automatically substituted for a wrapper script that "
              "enters your working directory and optionally starts one of your Custom Commands.",
              "$COMMAND")),
        "gitPath": "git",
        "ownAskpass": _("Have OpenSSH ask for passphrases via {app}", app=APP_DISPLAY_NAME),
        "ownAskpass_help": paragraphs(
            _("Tick this to have OpenSSH use {app} to ask for passphrases."),
            _("Untick this if you’ve set up another program in the {0} environment variable (such as {1}).", tquo("SSH_ASKPASS"), tquo("ksshaskpass"))),
        "ownSshAgent": "ssh-agent",
        "ownSshAgent_false": _("Use ssh-agent provided by the system") + " " + "" if sshAuthSock else _("(not detected)"),
        "ownSshAgent_true": _("Have {app} manage its own ssh-agent", app=APP_DISPLAY_NAME),
        "ownSshAgent_help": paragraphs(
            _("“ssh-agent” can save your SSH credentials so you don’t have to retype the same passphrase over and over. "
              "Some Linux distributions set up an ssh-agent for you."),
            _("You can also have {app} start its own instance of ssh-agent "
              "for the duration of your session and have it remember passphrases."),
            sshAuthSockHelp,
        ),
        "lfsAware": _("Parse LFS pointers"),
        "lfsAware_help": paragraphs(
            _("Tick this to display the real contents from LFS files."),
            _("Untick to display the raw text in LFS pointers."),
        ),

        "mouseShortcuts": _("Mouse Shortcuts"),
        "mouseShortcuts_HEADER": _("Tip: If your mouse has side buttons, you can use them to navigate back/forward in the repo."),
        "tabBarClicks": _("Repository tabs:"),
        "doubleClickTabBar": _("Double-click tab"),
        "middleClickTabBar": _("Middle-click tab"),
        "fileListClicks": _("File lists:"),
        "doubleClickFileList": _("Double-click file"),
        "middleClickFileList": _("Middle-click file"),
        "diffViewClicks": _("Diff view:"),
        "middleClickStageLines": _("Middle-click selection"),
        "middleClickStageLines_true": _("Stage/unstage selected lines"),
        "middleClickStageLines_false": _("Do nothing"),

        "userCommands_guide": _userCommandsGuide(),
        "commands": "",
        "confirmCommands": _("Ask for confirmation before running any command"),
        "confirmCommands_help": _(
            "If you untick this, you can still force a prompt to appear for "
            "specific commands by prepending them with {0}. For example: {1}",
            tquo(f"<tt>{UserCommand.AlwaysConfirmPrefix}</tt>"),
            "<pre>?git stash</pre>"),
    }


def _timeFormatGuide() -> str:
    from gitfourchette.qt import QLocale, QDateTime, QDate, QTime

    locale = QLocale()
    firstDay = QDateTime(QDate(2000, 1, 1), QTime(0, 0))
    lastDay = QDateTime(QDate(2099, 12, 31), QTime(23, 59, 59))
    monday = QDateTime(QDate(2024, 12, 23), QTime(12, 0))
    sunday = QDateTime(QDate(2024, 12, 29), QTime(12, 0))

    def row(fmt: str, caption="", date1: QDateTime | None = firstDay, date2=lastDay):
        sample = ""
        if date1 is not None:
            f1 = locale.toString(date1, fmt)
            f2 = locale.toString(date2, fmt)
            sample = f1 + "–" + f2
            if caption:
                sample = f", {sample}"
        return f"\n<code>{fmt:>4} </code> {caption}{sample}"

    return (
            "<html style='white-space: pre'>"
            + _p("date/time formats", "Available formats:")
            + "<p>"
            + row("yy", _("year"))
            + row("yyyy", _("year") + f", {QDate.currentDate().year()}", None)
            + "</p><p>"
            + row("M", _("month"))
            + row("MM", _("month"))
            + row("MMM")
            + row("MMMM")
            + "</p><p>"
            + row("d", _("day"))
            + row("dd", _("day"))
            + row("ddd", "", monday, sunday)
            + row("dddd", "", monday, sunday)
            + "</p><p>"
            + row("h", _("hour") + ", 0–23/1–12", None)
            + row("hh", _("hour") + ", 00–23/01–12", None)
            + row("mm", _("minute"))
            + row("ss", _("second"))
            + "</p><p>"
            + row("a")
            + row("A")
            + "</p>")


def _gpgStatusReferenceTable() -> str:
    from gitfourchette.repomodel import GpgStatus

    return _tokenReferenceTable({
        GpgStatus.Pending.iconHtml(): _("Verification pending"),
        GpgStatus.CantCheck.iconHtml(): _("Verification failed (e.g. missing key)"),
        GpgStatus.GoodTrusted.iconHtml(): _("Good signature; Key trusted"),
        GpgStatus.GoodUntrusted.iconHtml(): _("Good signature; Key not fully trusted"),
        GpgStatus.ExpiredSig.iconHtml(): _("Good signature; Key or signature expired"),
        GpgStatus.Bad.iconHtml(): _("Key revoked or bad signature"),
    })


def _tokenReferenceTable(table) -> str:
    markup = "<table>"
    for token, caption in table.items():
        markup += f"<tr><td><b>{token} </b></td><td> {caption}</td></tr>\n"
    markup += "</table>"
    return markup


def _userCommandsGuide() -> str:
    from gitfourchette.toolbox.textutils import paragraphs, linkify, stripAccelerators, tquo
    from gitfourchette.exttools.usercommand import UserCommand

    def cmdName(s: str):
        s = re.sub(r"&(.)", r"<u>&amp;\1</u>", s)
        return f"<com># {s}</com>"

    markup = textwrap.dedent("""\
    <style>
    body {background-color: palette(window); }
    pre { font-size: small; white-space: pre-wrap; margin: 0px; margin-left: 16px; }
    com { color: gray; font-style: italic; }
    tok { font-weight: bold; }
    </style><body>""")
    markup += paragraphs(_("Feel free to copy the sample below and paste it into the text box. "
                           "Then, click OK, and explore {menu} in the menu bar.",
                           menu=tquo(stripAccelerators(_("&Commands")))))
    markup += textwrap.dedent(f"""\
    <pre>
    git rebase -i <tok>$COMMIT</tok>   {cmdName(_("&Interactive Rebase"))}
    git rebase --continue   {cmdName(_("&Continue Rebase"))}
    <tok>?</tok> git rebase --abort    {cmdName(_("&Abort Rebase"))}
    git diff <tok>$COMMIT</tok> HEAD   {cmdName(_("Diff Commit With &HEAD"))}
    </pre>""")
    markup += paragraphs(_("You may use the following placeholders in your commands:"))
    markup += _tokenReferenceTable(UserCommand.tokenHelpTable())
    markup += paragraphs(linkify(_("For advanced usage tips, please visit [the user’s guide]."
                                   ), _userCommandsGuideUrl))
    return markup



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import math
import traceback
from contextlib import suppress
from dataclasses import dataclass

from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.graphview.commitlogmodel import CommitLogModel, SpecialRow, CommitToolTipZone
from gitfourchette.graphview.commitinfosearch import CommitInfoSearch
from gitfourchette.graphview.graphpaint import paintGraphFrame
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID, UC_FAKEREF, RepoModel, GpgStatus
from gitfourchette.toolbox import *


@dataclass
class RefBox:
    prefix: str
    icon: str = ""
    color: QColor = None
    keepPrefix: bool = False
    iconWidth: int = 16


REFBOXES = [
    RefBox(RefPrefix.REMOTES, "git-remote", QColor(Qt.GlobalColor.darkCyan)),
    RefBox(RefPrefix.TAGS, "git-tag", QColor(Qt.GlobalColor.darkYellow)),
    RefBox(RefPrefix.HEADS, "git-branch", QColor(Qt.GlobalColor.darkMagenta)),

    # detached HEAD as returned by Repo.map_commits_to_refs
    RefBox("HEAD", "git-head-detached", QColor(Qt.GlobalColor.darkRed), keepPrefix=True),

    # Working Directory
    RefBox(UC_FAKEREF, "git-workdir", QColor("#808080")),

    # Mounted
    RefBox("FAKEREF_FUSEMOUNT", "git-mount", QColor(Qt.GlobalColor.gray)),

    # Commit comparison
    RefBox("FAKEREF_COMPAREA", "compare-a", QColor(Qt.GlobalColor.red), iconWidth=48),
    RefBox("FAKEREF_COMPAREB", "compare-b", QColor(Qt.GlobalColor.blue), iconWidth=48),

    # Fallback
    RefBox("", "hint", QColor(Qt.GlobalColor.gray), keepPrefix=True)
]


ELISION = " […]"
ELISION_LENGTH = len(ELISION)


MAX_AUTHOR_CHARS = {
    AuthorDisplayStyle.Initials: 7,
    AuthorDisplayStyle.FullName: 20,
    AuthorDisplayStyle.FullEmail: 24,
}


XMARGIN = 4
XSPACING = 6

NARROW_WIDTH = (500, 750)


class CommitLogDelegate(QStyledItemDelegate):
    requestSignatureVerification = Signal(Oid)

    def __init__(
            self,
            repoModel: RepoModel,
            infoSearch: CommitInfoSearch | None = None,
            parent: QWidget | None = None,
    ):
        super().__init__(parent)

        self.repoModel = repoModel
        self.infoSearch = infoSearch

        self.mustRefreshMetrics = True
        self.hashCharWidth = 0
        self.dateMaxWidth = 0
        self.authorMaxWidth = 0
        self.activeCommitFont = QFont()
        self.uncommittedFont = QFont()
        self.refboxFont = QFont()
        self.homeRefboxFont = QFont()

        self.mounts = GFApplication.instance().mountManager

        self._transientToolTipZones: list[CommitToolTipZone] | None = None

    def prepareForDeletion(self):
        del self.repoModel
        del self.infoSearch
        del self.mounts

    def newToolTipZone(self, zone: CommitToolTipZone):
        if self._transientToolTipZones is None:
            self._transientToolTipZones = []
        self._transientToolTipZones.append(zone)

    # --------------------------------------------------------------------------
    # Metrics

    def invalidateMetrics(self):
        self.mustRefreshMetrics = True

    def refreshMetrics(self, option: QStyleOptionViewItem):
        if not self.mustRefreshMetrics:
            return

        self.mustRefreshMetrics = False

        self.hashCharWidth = max(option.fontMetrics.horizontalAdvance(c) for c in "0123456789abcdef")

        self.activeCommitFont = QFont(option.font)
        self.activeCommitFont.setBold(True)

        self.uncommittedFont = QFont(option.font)
        self.uncommittedFont.setItalic(True)

        self.refboxFont = QFont(option.font)

        self.homeRefboxFont = QFont(self.refboxFont)
        self.homeRefboxFont.setWeight(QFont.Weight.Bold)

        wideDate = QDateTime.fromString("2999-12-25T23:59:59.999", Qt.DateFormat.ISODate)
        dateText = option.locale.toString(wideDate, settings.prefs.shortTimeFormat)
        if settings.prefs.authorDiffAsterisk:
            dateText += "*"
        self.dateMaxWidth = QFontMetrics(self.activeCommitFont).horizontalAdvance(dateText + " ")
        self.dateMaxWidth = int(self.dateMaxWidth)  # make sure it's an int for pyqt5 compat

        self.authorMaxWidth = self.hashCharWidth * MAX_AUTHOR_CHARS.get(settings.prefs.authorDisplayStyle, 16)

    # --------------------------------------------------------------------------
    # Qt callbacks

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex) -> QSize:
        mult = settings.prefs.graphRowHeight
        r = super().sizeHint(option, index)
        r.setHeight(option.fontMetrics.height() * mult // 100)
        return r

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex, fillBackground=True):
        painter.save()
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        try:
            assert self._transientToolTipZones is None
            self._paint(painter, option, index, fillBackground)
        except Exception as exc:  # pragma: no cover
            painter.restore()
            painter.save()
            self._paintError(painter, option, index, exc)
        finally:
            self._transientToolTipZones = None
        painter.restore()

    # --------------------------------------------------------------------------
    # Paint implementation

    def _paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex, fillBackground: bool):
        assert index.isValid()

        isActive = bool(option.state & QStyle.StateFlag.State_Active)
        isSelected = bool(option.state & QStyle.StateFlag.State_Selected)
        colorGroup = QPalette.ColorGroup.Active if isActive else QPalette.ColorGroup.Inactive
        palette: QPalette = option.palette

        # Draw default background
        if fillBackground:
            style = option.widget.style()
            style.drawControl(QStyle.ControlElement.CE_ItemViewItem, option, painter, option.widget)

        if isSelected:
            painter.setPen(palette.color(colorGroup, QPalette.ColorRole.HighlightedText))

        # Get metrics of '0' before setting a custom font,
        # so that alignments are consistent in all commits regardless of bold or italic.
        self.refreshMetrics(option)

        # Get the commit
        # special: SpecialRow = index.data(CommitLogModel.Role.SpecialRow)
        oid = index.data(CommitLogModel.Role.Oid)
        if oid is not None and oid != UC_FAKEID:
            commit = self.repoModel.repo.peel_commit(oid)
        else:
            commit = None

        # Set up rect
        rect = QRect(option.rect)
        rect.setLeft(rect.left() + XMARGIN)
        rect.setRight(rect.right() - XMARGIN)
        fullWidth = rect.width()

        # Compute column bounds
        authorWidth = self.authorMaxWidth
        dateWidth = self.dateMaxWidth
        if fullWidth < NARROW_WIDTH[0] or not oid:
            authorWidth = 0
            dateWidth = 0
        elif fullWidth <= NARROW_WIDTH[1]:
            authorWidth = int(lerp(authorWidth/2, authorWidth, rect.width(), NARROW_WIDTH[0], NARROW_WIDTH[1]))
        leftBoundDate = rect.right() - dateWidth
        leftBoundName = leftBoundDate - authorWidth
        rightBound = rect.right()
        tabBound = rect.right() - authorWidth - dateWidth

        # Reserve rightmost column
        rect.setRight(leftBoundName - XMARGIN)

        # Set font
        if self.isBold(oid):
            painter.setFont(self.activeCommitFont)
        elif not oid:
            painter.setFont(self.uncommittedFont)

        # ...Left-to-right zones...

        # Hash
        painter.save()
        if not isSelected:  # use muted color for hash if not selected
            painter.setPen(palette.color(colorGroup, QPalette.ColorRole.PlaceholderText))
        self._paintHash(painter, rect, oid)
        painter.restore()

        # Private
        self.paintPrivate(painter, option, index, rect, oid)

        # Use muted color from here on out for foreign commits (unless selected)
        if not isSelected and self.isDim(oid):
            painter.setPen(Qt.GlobalColor.gray)

        # Message
        if commit is not None:
            self._paintCommitMessage(painter, rect, commit)
        else:
            special: SpecialRow = index.data(CommitLogModel.Role.SpecialRow)
            self._paintSpecialMessage(painter, rect, special)

        # Pathspec match
        if (oid is not None
                and self.repoModel.commitPathspecFilter.isReady()
                and oid in self.repoModel.commitPathspecFilter.matchingIds):
            self._paintRefspecMatch(painter, rect, fullWidth // 8)

        # ...Jump to rightmost column...

        # Author
        if authorWidth != 0 and commit:
            rect.setLeft(tabBound)
            rect.setRight(leftBoundDate - XMARGIN)
            self._paintAuthor(painter, rect, commit)

        # Date
        if dateWidth != 0 and commit:
            rect.setLeft(leftBoundDate)
            rect.setRight(rightBound)
            self._paintDate(painter, rect, commit)

        # Set author/date tooltip zone
        if authorWidth != 0 or dateWidth != 0:
            self.newToolTipZone(CommitToolTipZone(leftBoundName, rightBound, "author"))

        # Tooltip metrics
        # Block model signals to update it - otherwise QComboBox will constantly redraw itself
        model = index.model()
        with QSignalBlockerContext(model):
            model.setData(index, leftBoundName if authorWidth != 0 else -1, CommitLogModel.Role.AuthorColumnX)
            model.setData(index, self._transientToolTipZones, CommitLogModel.Role.ToolTipZones)

        # Flush temp mouse zones
        self._transientToolTipZones = None

    # --------------------------------------------------------------------------
    # Paint blocks

    def _paintHash(self, painter: QPainter, rect: QRect, oid: Oid | None):
        hcw = self.hashCharWidth
        hashText = shortHash(oid) if oid else ("·" * settings.prefs.shortHashChars)

        charRect = QRect(rect)
        charRect.setWidth(hcw)

        for hashChar in hashText:
            painter.drawText(charRect, Qt.AlignmentFlag.AlignCenter, hashChar)
            charRect.translate(hcw, 0)

        # Highlight searched hash
        if (self.infoSearch is not None
                and self.infoSearch.likelyHash
                and (term := self.infoSearch.term())
                and oid is not None
                and str(oid).startswith(term)):
            x1 = 0
            x2 = min(len(hashText), len(term)) * hcw
            SearchBar.highlightNeedle(painter, rect, hashText, 0, len(term), x1, x2)

        rect.setLeft(charRect.right())

    def _paintSpecialMessage(self, painter: QPainter, rect: QRect, special: SpecialRow):
        if special == SpecialRow.UncommittedChanges:
            text = self.uncommittedChangesMessage()
        elif special == SpecialRow.EndOfShallowHistory:
            text = _("Shallow clone – End of commit history")
        elif special == SpecialRow.TruncatedHistory:
            if self.repoModel.hiddenCommits and self.repoModel.hiddenRefs:
                text = _("History truncated to {0} commits (including hidden branches)")
            else:
                text = _("History truncated to {0} commits")
            text = text.format(QLocale().toString(self.repoModel.numRealCommits))
        else:
            raise NotImplementedError(f"*** Unsupported special row {special}")

        text = painter.fontMetrics().elidedText(text, Qt.TextElideMode.ElideRight, rect.width())
        painter.drawText(rect, Qt.AlignmentFlag.AlignVCenter, text)

    def _paintCommitMessage(self, painter: QPainter, rect: QRect, commit: Commit):
        fullText = commit.message
        text, _contd = messageSummary(fullText, ELISION)

        text = painter.fontMetrics().elidedText(text, Qt.TextElideMode.ElideRight, rect.width())
        painter.drawText(rect, Qt.AlignmentFlag.AlignVCenter, text)

        if len(text) == 0 or text.endswith(("…", ELISION)):
            self.newToolTipZone(CommitToolTipZone(rect.left(), rect.right(), "message"))

        # Highlight search term
        if (text
                and self.infoSearch is not None
                and (searchTerm := self.infoSearch.term())
                and searchTerm in fullText.lower()):
            needlePos = text.lower().find(searchTerm)
            if needlePos < 0:
                needlePos = len(text) - ELISION_LENGTH
                needleLen = ELISION_LENGTH
            else:
                needleLen = len(searchTerm)
            SearchBar.highlightNeedle(painter, rect, text, needlePos, needleLen)

    def _paintRefspecMatch(self, painter: QPainter, rect: QRect, maxWidth: int):
        metrics = painter.fontMetrics()
        bleed, iconWidth = 8, 16

        needle = self.repoModel.commitPathspecFilter.needle
        needle = metrics.elidedText(needle, Qt.TextElideMode.ElideRight, maxWidth)

        needleRect = QRect(rect)
        needleRect.setRight(rect.right() - bleed)
        needleRect.setLeft(needleRect.right() - metrics.horizontalAdvance(needle))

        self.newToolTipZone(CommitToolTipZone(needleRect.left(), needleRect.right(), "pathspec"))

        SearchBar.highlightNeedle(painter, needleRect, needle, lBleed=bleed + iconWidth, rBleed=bleed)

        needleRect.adjust(-iconWidth, 0, 0, 0)
        needleRect.setWidth(iconWidth)
        stockIcon("magnifying-glass", "gray=black").paint(painter, needleRect)

    def _paintAuthor(self, painter: QPainter, rect: QRect, commit: Commit):
        assert commit
        author = commit.author
        authorText = abbreviatePerson(author, settings.prefs.authorDisplayStyle)

        if settings.prefs.authorDiffAsterisk and author.email != commit.committer.email:
            authorText += "*"

        gpgStatus, _gpgKeyInfo = self.repoModel.getCachedGpgStatus(commit)

        if gpgStatus == GpgStatus.Pending and settings.prefs.verifyGpgOnTheFly:
            self.requestSignatureVerification.emit(commit.id)

        # Draw seal for signed commits
        if gpgStatus > GpgStatus.Pending or (settings.prefs.verifyGpgOnTheFly and gpgStatus >= GpgStatus.Pending):
            sealRect = QRect(rect)
            sealRect.setRight(sealRect.left() + 16)
            icon = stockIcon(gpgStatus.iconName())
            icon.paint(painter, sealRect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
            rect.setLeft(sealRect.right() + 4)

        FittedText.draw(painter, rect, Qt.AlignmentFlag.AlignVCenter, authorText, minStretch=QFont.Stretch.ExtraCondensed)

        # Highlight searched author
        if self.infoSearch is not None and (searchTerm := self.infoSearch.term()):
            needlePos = authorText.lower().find(searchTerm)
            if needlePos >= 0:
                SearchBar.highlightNeedle(painter, rect, authorText, needlePos, len(searchTerm))

    def _paintDate(self, painter: QPainter, rect: QRect, commit: Commit):
        author = commit.author
        dateText = signatureDateFormat(author, settings.prefs.shortTimeFormat, localTime=True)

        if settings.prefs.authorDiffAsterisk and author.time != commit.committer.time:
            dateText += "*"

        displayText = painter.fontMetrics().elidedText(dateText, Qt.TextElideMode.ElideRight, rect.width())
        painter.drawText(rect, Qt.AlignmentFlag.AlignVCenter, displayText)

    # --------------------------------------------------------------------------
    # Refbox painting

    def _paintRefboxes(self, painter: QPainter, rect: QRect, refs: list[str]):
        repoModel = self.repoModel
        homeBranch = RefPrefix.HEADS + repoModel.homeBranch
        xMax = painter.clipBoundingRect().right()

        # Group refs in clusters (branches with same upstream)
        clusters: dict[str, list[str]] = {}
        nonLooseRefs = set()
        for refName in refs:
            # Skip refboxes for hidden refs
            if refName in repoModel.hiddenRefs:
                continue

            # Look for local branches
            if not refName.startswith(RefPrefix.HEADS):
                continue
            localName = refName.removeprefix(RefPrefix.HEADS)

            # Find the upstream for this local branch
            try:
                upstreamShorthand = repoModel.upstreams[localName]
                assert not upstreamShorthand.startswith(RefPrefix.REMOTES)
                upstreamRef = RefPrefix.REMOTES + upstreamShorthand
            except KeyError:
                continue

            # Don't create a cluster if the upstream isn't on the same row
            if upstreamRef not in refs:
                continue

            # Don't create a cluster if the upstream is hidden
            if upstreamRef in repoModel.hiddenRefs:
                continue

            # Append to the cluster
            try:
                clusters[upstreamRef].append(refName)
            except KeyError:
                clusters[upstreamRef] = [refName]

            nonLooseRefs.add(upstreamRef)
            nonLooseRefs.add(refName)

        # Draw clusters first
        for upstreamRef, localRefList in clusters.items():
            # See if we can omit the name of the remote branch
            if repoModel.singleRemote and len(localRefList) == 1:
                assert upstreamRef.startswith(RefPrefix.REMOTES)
                upstreamShorthand = upstreamRef.removeprefix(RefPrefix.REMOTES)
                _remoteName, remoteBranchName = split_remote_branch_shorthand(upstreamShorthand)
                _localBranchPrefix, localBranchName = RefPrefix.split(localRefList[0])
                omitRemoteName = remoteBranchName == localBranchName
            else:
                omitRemoteName = False

            # Draw local branches
            for i, localRef in enumerate(localRefList):
                self._paintRefbox(painter, rect, localRef, clipLeft=i != 0, clipRight=True, isHome=localRef == homeBranch)

            # Draw upstream at end of cluster
            self._paintRefbox(painter, rect, upstreamRef, clipLeft=True, forceOmitName=omitRemoteName)

            if rect.left() >= xMax:
                return

        # Draw loose refs
        for refName in refs:
            # Skip refboxes for hidden refs (except tags and special refs)
            if (refName in repoModel.hiddenRefs
                    and refName.startswith("refs/")
                    and not refName.startswith(RefPrefix.TAGS)):
                continue

            # Skip clustered refs we've drawn above
            if refName in nonLooseRefs:
                continue

            self._paintRefbox(painter, rect, refName, isHome=refName == homeBranch)

            if rect.left() >= xMax:
                return

    def _paintRefbox(
            self,
            painter: QPainter,
            rect: QRect,
            refName: str,
            isHome: bool = False,
            clipLeft: bool = False,
            clipRight: bool = False,
            forceOmitName: bool = False,
            forceToolTip: str | None = "",
    ):
        if refName == 'HEAD' and not self.repoModel.headIsDetached:
            return

        refboxDef = next(d for d in REFBOXES if refName.startswith(d.prefix))

        penColor = painter.pen().color()

        if forceOmitName:
            text = ""
        elif not refboxDef.keepPrefix:
            text = refName.removeprefix(refboxDef.prefix)
        else:
            text = refName
        color = refboxDef.color or penColor
        bgColor = QColor(color)  # modify copy
        iconName = refboxDef.icon

        # Omit remote name if there's a single remote
        if refboxDef.prefix == RefPrefix.REMOTES and self.repoModel.singleRemote:
            text = text.split('/', 1)[-1]

        dark = penColor.lightnessF() > .5
        if dark:
            color = color.lighter(300)
            bgColor.setAlphaF(.5)
        else:
            bgColor.setAlphaF(.066)

        if isHome:
            font = self.homeRefboxFont
            iconName = "git-head"
        elif refName == 'HEAD' and self.repoModel.headIsDetached:
            text = _("Detached HEAD")
            font = self.homeRefboxFont
        else:
            font = self.refboxFont

        painter.setFont(font)
        painter.setPen(color)

        rrRadius = 4  # Rounded Rectangle radius
        lPadding = 4  # Left padding
        rPadding = 4  # Right padding
        vMargin = max(0, math.ceil((rect.height() - 16) / 4))  # Vertical margin

        if iconName:
            lPadding -= 1

        # Determine max width
        maxWidth = int(settings.prefs.refBoxMaxWidth)
        remainingWidth = rect.width()
        if remainingWidth < 150:  # Super cramped
            maxWidth = 0  # Draw icon only
        maxWidth = min(remainingWidth, maxWidth)

        # Text-only refbox: show text regardless of the user's preference
        if not iconName and text:
            maxWidth = max(maxWidth, remainingWidth)

        # Draw text
        if text and maxWidth != 0:
            text, fittedFont, textWidth = FittedText.fit(
                font, maxWidth, text, Qt.TextElideMode.ElideMiddle, limit=QFont.Stretch.Condensed)
        else:
            textWidth = -rPadding  # Negate rPadding

        lClip = 0
        rClip = 0
        if clipLeft:
            lPadding = 2 * lPadding
            lClip = rrRadius
        if clipRight:
            rPadding = 2 * rPadding + 2
            rClip = rrRadius

        if iconName:
            iconRect = QRect(rect)
            iconRect.adjust(lPadding, vMargin, 0, -vMargin)
            iconHeight = min(16, iconRect.height())
            iconWidth = round(refboxDef.iconWidth * iconHeight / 16)
            iconRect.setWidth(iconWidth)
            iconPadding = 2
        else:
            iconWidth = 0
            iconPadding = 0

        boxRect = QRect(rect)
        boxRect.setWidth(lPadding + iconWidth + iconPadding + textWidth + rPadding)

        frameRect = QRectF(boxRect)
        frameRect.adjust(0, vMargin, 0, -vMargin)
        frameRect.adjust(-lClip, 0, 0, 0)
        clipBox = frameRect.adjusted(lClip, 0, -rClip+1, 0)

        if lClip or rClip:
            painter.save()
            painter.setClipRect(clipBox)

        framePath = QPainterPath()
        framePath.addRoundedRect(frameRect.adjusted(.5, .5, .5, -.5),  # Snap to pixel grid
                                 rrRadius, rrRadius)

        painter.drawPath(framePath)
        painter.fillPath(framePath, bgColor)

        if iconName:
            icon = stockIcon(iconName, f"gray={color.name()}")
            icon.paint(painter, iconRect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)

        if text and textWidth > 0:
            textRect = QRect(boxRect)
            textRect.adjust(0, 0, -rPadding, 0)
            painter.setFont(fittedFont)
            painter.drawText(textRect, Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight, text)
            painter.setFont(font)

        # Reset clip rect
        if lClip or rClip:
            painter.restore()

        # Draw divider line
        if lClip:
            x  =  .5 + clipBox.left()
            yt =  .5 + frameRect.top()
            yb = -.5 + frameRect.bottom()
            ym =  .5 + int((yt+yb)/2)

            if rClip:
                # Simple divider
                divider = [QLineF(x, yt, x, yb)]
            else:
                # "Equals" sign punches through divider
                divider = [QLineF(x, yt, x, ym-3),
                           QLineF(x, ym+3, x, yb),
                           QLineF(x-2, ym-1, x+2, ym-1),
                           QLineF(x-2, ym+1, x+2, ym+1)]

            if PYSIDE6:
                painter.drawLines(divider)  # type: ignore[call-overload] # skip PySide6 for type checking
            else:
                painter.drawLines(*divider)

        # Append tooltip
        if forceToolTip is not None:
            toolTipText = forceToolTip or refName
            zone = CommitToolTipZone(rect.left(), boxRect.right(), "ref", toolTipText)
            self.newToolTipZone(zone)

        # Advance caller rectangle
        rect.setLeft(round(clipBox.right()) + (6 if not rClip else 0))

    # --------------------------------------------------------------------------
    # Error painting

    def _paintError(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex, exc: Exception):  # pragma: no cover
        """Last-resort row drawing routine used if _paint raises an exception."""

        # We want this to fail in unit tests.
        if APP_TESTMODE:
            raise exc

        text = "?" * 7
        with suppress(Exception):
            oid = index.data(CommitLogModel.Role.Oid)
            text = str(oid)[:7]
        with suppress(Exception):
            details = traceback.format_exception(exc.__class__, exc, exc.__traceback__)
            text += " - " + shortenTracebackPath(details[-2].splitlines(False)[0]) + ":: " + repr(exc)

        bg, fg = QColor(Qt.GlobalColor.white), QColor(Qt.GlobalColor.red)
        if option.state & QStyle.StateFlag.State_Selected:
            bg, fg = fg, bg

        painter.fillRect(option.rect, bg)
        painter.setPen(fg)
        painter.setFont(QFontDatabase.systemFont(QFontDatabase.SystemFont.SmallestReadableFont))
        painter.drawText(option.rect, Qt.AlignmentFlag.AlignVCenter, text)

    # --------------------------------------------------------------------------
    # To override

    def isBold(self, oid: Oid) -> bool:
        """Can be overridden"""
        return oid != NULL_OID and oid == self.repoModel.headCommitId

    def isDim(self, oid: Oid):
        """Can be overridden"""
        return oid != NULL_OID and oid in self.repoModel.foreignCommits

    def uncommittedChangesMessage(self) -> str:
        """Can be overridden"""
        summaryText = _("Working Directory") + " "
        # Append change count if available
        numChanges = self.repoModel.numUncommittedChanges
        if numChanges == 0:
            summaryText += _("(Clean)")
        elif numChanges > 0:
            summaryText += _n("({n} change)", "({n} changes)", numChanges)
        # Append draft message if any
        draftMessage = self.repoModel.prefs.draftCommitMessage
        if draftMessage:
            draftMessage = messageSummary(draftMessage)[0].strip()
            draftIntro = _("Commit draft:")
            summaryText += f" – {draftIntro} {tquo(draftMessage)}"
        return summaryText

    def paintPrivate(
            self,
            painter: QPainter,
            option: QStyleOptionViewItem,
            index: QModelIndex,
            rect: QRect,
            oid: Oid | None,
    ):
        """
        Draw widget-specific information inbetween the commit hash and message.
        By default, draws the graph and refboxes. Can be overridden.
        """

        # ------ Graph
        if oid is not None and not self.repoModel.commitPathspecFilter.wantFilter():
            graphRect = QRect(rect)
            paintGraphFrame(painter, graphRect, oid, self.repoModel.graph, self.repoModel.hiddenCommits)
            rect.setLeft(graphRect.right())

        # ------ Begin refboxes
        painter.save()
        painter.setClipRect(rect)

        # ------ A/B icon
        abSide = index.data(CommitLogModel.Role.ComparisonSide)
        if abSide:
            tt = _("{0} side in the current comparison between two commits", tquo(abSide))
            self._paintRefbox(painter, rect, f"FAKEREF_COMPARE{abSide}", forceToolTip=tt)

        # ------ Mount icon
        if self.mounts.isMounted(oid):
            tt = _("This commit is currently mounted as a folder.")
            self._paintRefbox(painter, rect, "FAKEREF_FUSEMOUNT", forceToolTip=tt)

        # ------ Actual refs
        refsHere = self.repoModel.refsAt.get(oid, None)
        if refsHere:
            self._paintRefboxes(painter, rect, refsHere)

        # ------ End refboxes
        painter.restore()




# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import copy
import gc
import logging
import os
import re
from collections.abc import Sequence, Callable
from contextlib import suppress
from pathlib import Path
from typing import Literal

from gitfourchette import settings
from gitfourchette import tasks
from gitfourchette.application import GFApplication
from gitfourchette.codeview.codeview import CodeView
from gitfourchette.dropzone import DropAction, DropZone
from gitfourchette.exttools.toolprocess import ToolProcess, setUpToolCommand
from gitfourchette.exttools.usercommand import UserCommand
from gitfourchette.forms.aboutdialog import AboutDialog
from gitfourchette.forms.clonedialog import CloneDialog
from gitfourchette.forms.maintoolbar import MainToolBar
from gitfourchette.forms.repostub import RepoStub
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.forms.textinputdialog import TextInputDialog
from gitfourchette.forms.welcomewidget import WelcomeWidget
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext, NavFlags
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repowidget import RepoWidget
from gitfourchette.settings import PrefEffects, TabBarClick, getExternalEditorName
from gitfourchette.syntax import LexJobCache
from gitfourchette.tasks import TaskBook, RepoTaskRunner
from gitfourchette.tasks.newrepotasks import NewRepo
from gitfourchette.toolbox import *
from gitfourchette.trash import Trash

logger = logging.getLogger(__name__)

USERS_GUIDE_URL = "https://gitfourchette.org/guide"


class NoRepoWidgetError(Exception):
    pass


class MainWindow(QMainWindow):
    welcomeStack: QStackedWidget
    welcomeWidget: WelcomeWidget
    tabs: QTabWidget2

    recentMenu: QMenu
    showStatusBarAction: QAction
    showMenuBarAction: QAction

    def __init__(self) -> None:
        super().__init__()

        self.welcomeStack = QStackedWidget(self)
        self.setCentralWidget(self.welcomeStack)

        self.setObjectName("GFMainWindow")

        self.setWindowTitle(qAppName())

        initialSize = .75 * QApplication.primaryScreen().availableSize()
        self.resize(initialSize)

        self.tabs = QTabWidget2(self)
        self.tabs.currentWidgetChanged.connect(self.onTabCurrentWidgetChanged)
        self.tabs.tabCloseRequested.connect(self.closeTab)
        self.tabs.tabContextMenuRequested.connect(self.onTabContextMenu)
        self.tabs.tabMiddleClicked.connect(self.onTabMiddleClicked)
        self.tabs.tabDoubleClicked.connect(self.onTabDoubleClicked)

        self.welcomeWidget = WelcomeWidget(self)
        self.welcomeWidget.newRepo.connect(self.newRepo)
        self.welcomeWidget.openRepo.connect(self.openDialog)
        self.welcomeWidget.cloneRepo.connect(self.cloneDialog)

        self.welcomeStack.addWidget(self.welcomeWidget)
        self.welcomeStack.addWidget(self.tabs)
        self.welcomeStack.setCurrentWidget(self.welcomeWidget)

        self.globalMenuBar = QMenuBar(self)
        self.globalMenuBar.setObjectName("GFMainMenuBar")
        self.setMenuBar(self.globalMenuBar)
        self.autoHideMenuBar = AutoHideMenuBar(self.globalMenuBar)

        self.statusBar2 = QStatusBar2(self)
        self.setStatusBar(self.statusBar2)

        self.mainToolBar = MainToolBar(self)
        self.addToolBar(self.mainToolBar)
        self.mainToolBar.openDialog.connect(self.openDialog)
        self.mainToolBar.openPrefs.connect(GFApplication.instance().openPrefsDialog)
        self.mainToolBar.reveal.connect(lambda: self.currentRepoWidget().openRepoFolder())
        self.mainToolBar.openTerminal.connect(lambda: self.currentRepoWidget().openTerminal())
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.PreventContextMenu)

        self.recentMenu = QMenu(self)
        self.recentMenu.setObjectName("RecentMenu")
        self.recentMenu.setToolTipsVisible(True)
        self.fillRecentMenu()

        self.welcomeWidget.ui.recentReposButton.setMenu(self.recentMenu)
        self.mainToolBar.recentAction.setMenu(self.recentMenu)

        self.fillGlobalMenuBar()

        self.setAcceptDrops(True)

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

        self.dropZone = DropZone(self)
        self.dropZone.setVisible(False)

    # -------------------------------------------------------------------------
    # RepoTask

    @property
    def taskRunner(self) -> RepoTaskRunner:
        """
        Return the foreground RepoWidget's RepoTaskRunner.
        This provides compatibility for TaskInvocation's RepoTaskRunner lookup.
        May raise NoRepoWidgetError, aborting the lookup.
        """
        return self.currentRepoWidget().taskRunner

    # -------------------------------------------------------------------------
    # Event handlers

    def onMouseSideButtonPressed(self, forward: bool) -> None:
        if not self.isActiveWindow():
            return
        with suppress(NoRepoWidgetError):
            repoWidget = self.currentRepoWidget()
            if forward:
                repoWidget.navigateForward()
            else:
                repoWidget.navigateBack()

    def onFileDraggedToDockIcon(self, path: str) -> None:
        outcome = self.getDropOutcomeFromLocalFilePath(path)
        self.handleDrop(*outcome)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key.Key_Alt and self.autoHideMenuBar.enabled:
            self.autoHideMenuBar.toggle()
        else:
            super().keyReleaseEvent(event)

    def dragEnterEvent(self, event: QDragEnterEvent) -> None:
        try:
            action, data = self.getDropOutcomeFromMimeData(event.mimeData())
            if action == DropAction.Deny and not data:
                event.setAccepted(False)
                return
            self.dropZone.install(action, data)
            if action == DropAction.Deny:
                event.setDropAction(Qt.DropAction.IgnoreAction)  # 'nope' cursor on KDE
            else:
                event.acceptProposedAction()
        except Exception:  # pragma: no cover - Don't let this crash the application
            logger.exception("dragEnterEvent failed")
            event.setAccepted(False)

    def dragLeaveEvent(self, event: QDragLeaveEvent) -> None:
        self.dropZone.setVisible(False)

    def dropEvent(self, event: QDropEvent) -> None:
        self.dropZone.setVisible(False)
        action, data = self.getDropOutcomeFromMimeData(event.mimeData())
        event.setAccepted(True)  # keep dragged item from coming back to cursor on macOS
        self.handleDrop(action, data)

    # -------------------------------------------------------------------------
    # Menu bar

    def fillGlobalMenuBar(self) -> None:
        menubar = self.globalMenuBar
        menuObjectNamePrefix = "MWMainMenu"

        # Delete old menu objects
        for m in menubar.findChildren(QMenu, options=Qt.FindChildOption.FindDirectChildrenOnly):
            if m.objectName().startswith(menuObjectNamePrefix):
                m.deleteLater()

        menubar.clear()

        # -------------------------------------------------------------
        # Set up root menus

        menuNames = {
            "File": _("&File"),
            "Edit": _("&Edit"),
            "View": _("&View"),
            "Repo": _("&Repo"),
            "Commands": _("&Commands"),
            "Mount": _("&Mount"),
            "Help": _("&Help"),
        }

        rootMenus: dict[str, QMenu] = {}
        for key, name in menuNames.items():
            menu = menubar.addMenu(name)
            menu.setObjectName(f"{menuObjectNamePrefix}{key}")
            menu.setToolTipsVisible(True)
            rootMenus[key] = menu

        fileMenu, editMenu, viewMenu, repoMenu, commandsMenu, mountMenu, helpMenu = iter(rootMenus.values())

        self.autoHideMenuBar.reconnectToMenus()

        # -------------------------------------------------------------

        ActionDef.addToQMenu(
            fileMenu,

            ActionDef(_("&New Repository…"), self.newRepo,
                      shortcuts=QKeySequence.StandardKey.New, icon="folder-new",
                      tip=_("Create an empty Git repo")),

            ActionDef(_("C&lone Repository…"), self.cloneDialog,
                      shortcuts="Ctrl+Shift+N", icon="folder-download",
                      tip=_("Download a Git repo and open it")),

            ActionDef.SEPARATOR,

            ActionDef(_("&Open Repository…"), self.openDialog,
                      shortcuts=QKeySequence.StandardKey.Open, icon="folder-open",
                      tip=_("Open a Git repo on your machine")),

            ActionDef(_("Open &Recent"),
                      icon="folder-open-recent",
                      tip=_("List of recently opened Git repos"),
                      submenu=self.recentMenu),

            ActionDef.SEPARATOR,

            TaskBook.action(self, tasks.ApplyPatchFile),
            TaskBook.action(self, tasks.ApplyPatchFileReverse),

            ActionDef.SEPARATOR,

            ActionDef(_("&Settings…"), GFApplication.instance().openPrefsDialog,
                      shortcuts=QKeySequence.StandardKey.Preferences, icon="configure",
                      menuRole=QAction.MenuRole.PreferencesRole,
                      tip=_("Configure {app}", app=qAppName())),

            TaskBook.action(self, tasks.SetUpGitIdentity, taskArgs=('', False)
                            ).replace(menuRole=QAction.MenuRole.ApplicationSpecificRole),

            ActionDef.SEPARATOR,

            ActionDef(_("&Close Tab"), self.dispatchCloseCommand,
                      shortcuts=QKeySequence.StandardKey.Close, icon="document-close",
                      tip=_("Close current repository tab")),

            ActionDef(_("&Quit"), self.close,
                      shortcuts=QKeySequence.StandardKey.Quit, icon="application-exit",
                      tip=_("Quit {app}", app=qAppName()),
                      menuRole=QAction.MenuRole.QuitRole),
        )

        # -------------------------------------------------------------

        ActionDef.addToQMenu(
            editMenu,

            ActionDef(_("&Find…"), lambda: self.dispatchSearchCommand(),
                      shortcuts=GlobalShortcuts.find, icon="edit-find",
                      tip=_("Search for a piece of text in commit messages, the current diff, or the name of a file")),

            ActionDef(_("Find Next"), lambda: self.dispatchSearchCommand(SearchBar.Op.Next),
                      shortcuts=GlobalShortcuts.findNext,
                      tip=_("Find next occurrence")),

            ActionDef(_("Find Previous"), lambda: self.dispatchSearchCommand(SearchBar.Op.Previous),
                      shortcuts=GlobalShortcuts.findPrevious,
                      tip=_("Find previous occurrence")),
        )

        # -------------------------------------------------------------

        ActionDef.addToQMenu(
            repoMenu,
            *RepoWidget.contextMenuItemsByProxy(self, self.currentRepoWidget),
        )

        # -------------------------------------------------------------

        ActionDef.addToQMenu(
            viewMenu,
            self.mainToolBar.toggleViewAction(),
            ActionDef(englishTitleCase(_("Show status bar")), self.toggleStatusBar, checkState=-1, objectName="ShowStatusBarAction"),
            ActionDef(englishTitleCase(_("Show menu bar")), self.toggleMenuBar, checkState=-1, objectName="ShowMenuBarAction"),
            ActionDef.SEPARATOR,
            TaskBook.action(self, tasks.JumpToUncommittedChanges, accel="U"),
            TaskBook.action(self, tasks.JumpToHEAD, accel="H"),
            ActionDef.SEPARATOR,
            ActionDef(_("Focus on Sidebar"), self.focusSidebar, shortcuts="Alt+1"),
            ActionDef(_("Focus on Commit Log"), self.focusGraph, shortcuts="Alt+2"),
            ActionDef(_("Focus on File List"), self.focusFiles, shortcuts="Alt+3"),
            ActionDef(_("Focus on Code View"), self.focusDiff, shortcuts="Alt+4"),
            ActionDef.SEPARATOR,
            ActionDef(_("Blame File…"), self.blameFile, icon=TaskBook.icons[tasks.OpenBlame], shortcuts=TaskBook.shortcuts[tasks.OpenBlame]),
            ActionDef(_("Next File"), self.nextFile, shortcuts="Ctrl+]"),
            ActionDef(_("Previous File"), self.previousFile, shortcuts="Ctrl+["),
            ActionDef.SEPARATOR,
            ActionDef(_("&Next Tab"), self.nextTab, shortcuts="Ctrl+Shift+]" if MACOS else "Ctrl+Tab"),
            ActionDef(_("&Previous Tab"), self.previousTab, shortcuts="Ctrl+Shift+[" if MACOS else "Ctrl+Shift+Tab"),
            ActionDef.SEPARATOR,
            TaskBook.action(self, tasks.JumpBack),
            TaskBook.action(self, tasks.JumpForward),
            ActionDef("Dump Nav Log", lambda: logger.info(self.currentRepoWidget().navHistory.getTextLog()), objectName="DumpNavLogAction"),
            ActionDef.SEPARATOR,
            ActionDef(
                _("&Refresh"),
                lambda: self.currentRepoWidget().refreshRepo(),
                shortcuts=GlobalShortcuts.refresh,
                icon="SP_BrowserReload",
                tip=_("Check for changes in the repo (on the local filesystem only – will not fetch remotes)"),
            ),
            ActionDef(
                _("Reloa&d"),
                lambda: self.currentRepoWidget().replaceWithStub(),
                shortcuts="Ctrl+F5",
                tip=_("Reopen the repo from scratch"),
            ),
        )

        self.showStatusBarAction = viewMenu.findChild(QAction, "ShowStatusBarAction")
        self.showMenuBarAction = viewMenu.findChild(QAction, "ShowMenuBarAction")
        self.showMenuBarAction.setVisible(not MACOS)
        viewMenu.findChild(QAction, "DumpNavLogAction").setVisible(APP_DEBUG)

        # -------------------------------------------------------------

        self.parseUserCommands()
        if self.userCommands:
            commandActions = [
                ActionDef.SEPARATOR if command.isSeparator
                else ActionDef(
                    command.menuTitle(),
                    lambda c=command: self.currentRepoWidget().executeUserCommand(c),
                    tip=command.menuToolTip(),
                    shortcuts=command.shortcut
                )
                for command in self.userCommands
            ]

            ActionDef.addToQMenu(
                commandsMenu,
                *commandActions,
                ActionDef.SEPARATOR,
                ActionDef(_("Edit Commands…"), icon="document-edit",
                          callback=lambda: GFApplication.instance().openPrefsDialog("commands")),
            )

            # Don't share commandsMenu with the terminal button: commandsMenu.aboutToShow
            # would fire via the terminal button's popup routine, causing AutoHideMenuBar to
            # show the entire menu bar.
            # Do share the actions themselves so that the keyboard shortcuts work.
            self.mainToolBar.setTerminalActions(commandsMenu.actions())
        else:
            commandsMenu.deleteLater()
            self.mainToolBar.setTerminalActions([])

        # -------------------------------------------------------------

        mountItems = GFApplication.instance().mountManager.makeMenu(self)
        if mountItems:
            ActionDef.addToQMenu(mountMenu, *mountItems)
        else:
            mountMenu.deleteLater()

        # -------------------------------------------------------------

        ActionDef.addToQMenu(
            helpMenu,

            ActionDef(
                _("&About {0}", qAppName()),
                lambda: AboutDialog.popUp(self),
                icon="gitfourchette",
                menuRole=QAction.MenuRole.AboutRole,),

            ActionDef(
                _("{0} User’s Guide", qAppName()),
                lambda: QDesktopServices.openUrl(QUrl(USERS_GUIDE_URL)),
                icon="help-contents"),

            ActionDef.SEPARATOR,

            ActionDef(
                _("Open Trash…"),
                self.openRescueFolder,
                icon="SP_TrashIcon",
                tip=_("Explore changes that you may have discarded by mistake")),

            ActionDef(
                _("Empty Trash…"),
                self.clearRescueFolder,
                tip=_("Delete all discarded changes from the trash folder")),
        )

    def fillRecentMenu(self) -> None:
        actions = []
        for path in settings.history.getRecentRepoPaths(settings.prefs.maxRecentRepos):
            caption = compactPath(path)
            nickname = settings.history.getRepoNickname(path, strict=True)
            if nickname:
                caption += f" ({tquo(nickname)})"

            caption = elide(caption, ems=60)  # don't let menus get absurdly wide

            openAction = ActionDef(
                escamp(caption),
                lambda p=path: self.openRepo(p, exactMatch=True),
                tip=path)
            actions.append(openAction)

        self.recentMenu.clear()
        ActionDef.addToQMenu(
            self.recentMenu,
            *actions,
            ActionDef.SEPARATOR,
            ActionDef(
                _("Clear List"), self.onClearRecentMenu, "edit-clear-history",
                tip=_("Clear the list of recently opened repositories"),
            ))

    def onClearRecentMenu(self) -> None:
        settings.history.clearRepoHistory()
        settings.history.write()
        self.fillRecentMenu()

    def showMenuBarHiddenWarning(self):
        return showInformation(
            self, _("Menu bar hidden"),
            _("The menu bar is now hidden. Press the Alt key to toggle it."))

    # -------------------------------------------------------------------------
    # Tabs

    def currentRepoWidget(self) -> RepoWidget:
        rw = self.tabs.currentWidget()
        if not isinstance(rw, RepoWidget):  # it might be a RepoStub
            raise NoRepoWidgetError()
        return rw

    def tabWidgetForWorkdirPath(self, workdir: str) -> RepoWidget | RepoStub | None:
        widget: RepoWidget | RepoStub
        for widget in self.tabs.widgets():
            assert isinstance(widget, RepoWidget | RepoStub)
            with suppress(FileNotFoundError):  # may be raised if workdir cannot be accessed
                if os.path.samefile(workdir, widget.workdir):
                    return widget
        return None

    def setWindowTitle(self, title: str) -> None:
        if APP_DEBUG:
            chain = ["DEBUG", str(os.getpid()), QT_BINDING]
            if APP_TESTMODE:
                chain.append("TESTMODE")
            if APP_NOTHREADS:
                chain.append("NOTHREADS")
            title = f"{title} ({' '.join(chain)})"

        super().setWindowTitle(title)

    def onTabCurrentWidgetChanged(self):
        self.mainToolBar.updateNavButtons()  # Kill back/forward arrows
        self.statusBar2.clearMessage()

        widget = self.tabs.currentWidget()

        # Switch to welcome widget if zero tabs
        if not widget:
            self.welcomeStack.setCurrentWidget(self.welcomeWidget)
            self.setWindowTitle(APP_DISPLAY_NAME)
            return

        self.welcomeStack.setCurrentWidget(self.tabs)  # Exit welcome widget
        self.setWindowTitle(widget.windowTitle())

        # If it's a RepoStub, load it if needed
        if not isinstance(widget, RepoWidget):
            assert isinstance(widget, RepoStub)
            if not widget.isPriming() and widget.willAutoLoad():
                widget.loadNow()
            return

        # We know it's a RepoWidget beyond this point
        assert isinstance(widget, RepoWidget)

        # Update back/forward buttons
        self.onRepoHistoryChanged(widget)

        widget.restoreSplitterStates()

        # Refresh the repo
        widget.refreshRepo()

    def generateTabContextMenu(self, i: int) -> QMenu:
        assert 0 <= i < self.tabs.count()

        widget: RepoWidget | RepoStub = self.tabs.widget(i)
        menu = QMenu(self)
        menu.setObjectName("MWRepoTabContextMenu")

        anyOtherLoadedTabs = any(tab is not widget and isinstance(tab, RepoWidget)
                                 for tab in self.tabs.widgets())

        ActionDef.addToQMenu(
            menu,
            ActionDef(_("Close Tab"), lambda: self.closeTab(i), shortcuts=QKeySequence.StandardKey.Close),
            ActionDef(_("Close Other Tabs"), lambda: self.closeOtherTabs(i), enabled=self.tabs.count() > 1),
            ActionDef(_("Unload Other Tabs"), lambda: self.unloadOtherTabs(i), enabled=self.tabs.count() > 1 and anyOtherLoadedTabs),
            ActionDef.SEPARATOR,
            *self.repolessActions(widget.workdir),
            ActionDef.SEPARATOR,
            ActionDef(_("Configure Tabs…"), lambda: GFApplication.instance().openPrefsDialog("tabCloseButton")),
        )

        return menu

    def onTabContextMenu(self, globalPoint: QPoint, i: int) -> None:
        if i < 0:  # Right mouse button released outside tabs
            return

        menu = self.generateTabContextMenu(i)
        menu.aboutToHide.connect(menu.deleteLater)
        menu.popup(globalPoint)

    def onTabMiddleClicked(self, i: int) -> None:
        self.onTabSpecialClick(i, "middle")

    def onTabDoubleClicked(self, i: int) -> None:
        self.onTabSpecialClick(i, "double")

    def onTabSpecialClick(self, i: int, click: Literal["middle", "double"]) -> None:
        if i < 0:
            return

        if click == "double":
            action = settings.prefs.doubleClickTabBar
        elif click == "middle":
            action = settings.prefs.middleClickTabBar
        else:
            raise NotImplementedError(f"unknown special click kind '{click}'")

        widget: RepoWidget | RepoStub = self.tabs.widget(i)

        if action == TabBarClick.Nothing:
            pass
        elif action == TabBarClick.Folder:
            openFolder(widget.workdir)
        elif action == TabBarClick.Close:
            self.closeTab(i)
        elif action == TabBarClick.Terminal:
            ToolProcess.startTerminal(self, widget.workdir)
        else:
            raise NotImplementedError(f"unknown special click action '{action}'")

    # -------------------------------------------------------------------------
    # Repo loading

    def openRepo(self, path: str, exactMatch=True) -> RepoWidget | RepoStub | None:
        try:
            rw = self._openRepo(path, exactMatch=exactMatch)
        except BaseException as exc:
            excMessageBox(
                exc,
                _("Open repository"),
                _("Couldn’t open the repository at {0}.", bquo(path)),
                parent=self,
                icon='warning')
            return None

        self.saveSession()

        # Return a concrete RepoWidget instead of a RepoStub if loading is already completed
        # (mostly for single-threaded unit tests that expect a deterministic chain of events).
        if RepoTaskRunner.ForceSerial:
            rw2 = self.tabWidgetForWorkdirPath(rw.workdir)
            if isinstance(rw2, RepoWidget):
                rw = rw2
            else:
                assert not APP_TESTMODE, "the RepoWidget isn't ready yet"

        return rw

    def _resolveWorkdir(self, path: str, exactMatch) -> tuple[str, RepoWidget | RepoStub | None]:
        # Make sure the path exists
        if not os.path.exists(path):
            raise FileNotFoundError(_("There’s nothing at this path."))

        # Resolve the workdir
        if not exactMatch:
            with RepoContext(path) as repo:
                if repo.is_bare:
                    raise NotImplementedError(_("Sorry, {app} doesn’t support bare repositories.", app=qAppName()))
                path = repo.workdir

        # Scan for an existing tab for this repo
        existingWidget = self.tabWidgetForWorkdirPath(path)
        if existingWidget is not None:
            return existingWidget.workdir, existingWidget

        # There's no widget for this workdir but it's a valid repo
        return path, None

    def _openRepo(self, path: str, foreground=True, tabIndex=-1, exactMatch=True, locator=NavLocator.Empty
                  ) -> RepoWidget | RepoStub:
        path, existingWidget = self._resolveWorkdir(path, exactMatch)

        # First check that we don't have a tab for this repo already
        if existingWidget is not None:
            existingWidget.overridePendingLocator(locator)
            self.tabs.setCurrentWidget(existingWidget)
            return existingWidget

        # Create a RepoStub
        stub = RepoStub(parent=self, workdir=path, locator=locator)

        # Create a tab
        with QSignalBlockerContext(self.tabs):
            title = escamp(stub.getTitle())
            tabIndex = self.tabs.insertTab(tabIndex, stub, title)
            self.tabs.setTabTooltip(tabIndex, compactPath(path))
            if foreground:
                self.tabs.setCurrentIndex(tabIndex)
        self.refreshAllTabTexts()

        # We've got at least one tab now, so switch away from WelcomeWidget
        assert self.tabs.count() > 0
        self.welcomeStack.setCurrentWidget(self.tabs)

        # Load repo now
        if foreground:
            self.setWindowTitle(stub.windowTitle())
            stub.loadNow()

        return stub

    def installRepoWidget(self, rw: RepoWidget, tabIndex: int) -> None:
        repoStub = self.tabs.widget(tabIndex)
        assert tabIndex >= 0, "stub to replace isn't in tabs"
        assert isinstance(repoStub, RepoStub), "yanked widget isn't RepoStub"

        rw.nameChange.connect(self.onRepoNameChanged)
        rw.requestAttention.connect(lambda: self.onRepoRequestsAttention(rw))
        rw.openRepo.connect(lambda path, locator: self.openRepoNextTo(rw, path, locator))
        rw.openPrefs.connect(GFApplication.instance().openPrefsDialog)
        rw.mustReplaceWithStub.connect(lambda stub: self.replaceRepoWidgetWithStub(rw, stub))

        rw.statusMessage.connect(self.statusBar2.showMessage)
        rw.busyMessage.connect(self.statusBar2.showBusyMessage)
        rw.clearStatus.connect(self.statusBar2.clearMessage)

        rw.historyChanged.connect(lambda: self.onRepoHistoryChanged(rw))
        rw.windowTitleChanged.connect(lambda: self.onRepoWindowTitleChanged(rw))

        self.tabs.swapWidget(tabIndex, rw)
        self.refreshAllTabTexts()

        repoStub.setParent(None)  # tabs don't deparent the widget
        repoStub.deleteLater()

    def replaceRepoWidgetWithStub(self, oldWidget: RepoWidget, stub: RepoStub) -> None:
        tabIndex = self.tabs.indexOf(oldWidget)
        assert tabIndex >= 0, "RepoWidget to replace isn't in tabs"
        self.tabs.swapWidget(tabIndex, stub)
        self.refreshAllTabTexts()

        oldWidget.setParent(None)  # tabs don't deparent the widget
        oldWidget.close()  # will call cleanup
        oldWidget.deleteLater()

    # -------------------------------------------------------------------------

    def onRegainForeground(self) -> None:
        if QGuiApplication.applicationState() != Qt.ApplicationState.ApplicationActive:
            return
        if not settings.prefs.autoRefresh:
            return
        with suppress(NoRepoWidgetError):
            self.currentRepoWidget().refreshRepo()

    def onRepoNameChanged(self) -> None:
        self.refreshAllTabTexts()
        self.fillRecentMenu()

    def onRepoWindowTitleChanged(self, rw: RepoWidget) -> None:
        if rw.isVisible():
            self.setWindowTitle(rw.windowTitle())

    def onRepoHistoryChanged(self, rw: RepoWidget) -> None:
        if rw.isVisible():
            self.mainToolBar.updateNavButtons(rw.navHistory.canGoBack(), rw.navHistory.canGoForward())

    def onRepoRequestsAttention(self, rw: RepoWidget) -> None:
        i = self.tabs.indexOf(rw)
        self.tabs.requestAttention(i)

    # -------------------------------------------------------------------------
    # View menu

    def toggleStatusBar(self) -> None:
        GFApplication.applyPrefs(showStatusBar=not settings.prefs.showStatusBar)

    def toggleMenuBar(self) -> None:
        GFApplication.applyPrefs(showMenuBar=not settings.prefs.showMenuBar)

    def selectUncommittedChanges(self) -> None:
        self.currentRepoWidget().jump(NavLocator.inWorkdir())

    def selectHead(self) -> None:
        self.currentRepoWidget().jump(NavLocator.inRef("HEAD"))

    def focusSidebar(self) -> None:
        self.currentRepoWidget().sidebar.setFocus()

    def focusGraph(self) -> None:
        self.currentRepoWidget().graphView.setFocus()

    def focusFiles(self) -> None:
        rw = self.currentRepoWidget()
        context = rw.navLocator.context
        if context == NavContext.COMMITTED:
            rw.committedFiles.setFocus()
        else:
            target = rw.stagedFiles if context == NavContext.STAGED else rw.dirtyFiles
            fallback = rw.dirtyFiles if context == NavContext.STAGED else rw.stagedFiles
            if not target.isEmpty() or fallback.isEmpty():
                target.setFocus()
            else:
                fallback.setFocus()

    def focusDiff(self) -> None:
        rw = self.currentRepoWidget()
        if rw.specialDiffView.isVisibleTo(rw):
            rw.specialDiffView.setFocus()
        elif rw.conflictView.isVisibleTo(rw):
            rw.conflictView.setFocus()
        else:
            rw.diffView.setFocus()

    def nextFile(self) -> None:
        self.currentRepoWidget().diffArea.selectNextFile(True)

    def previousFile(self) -> None:
        self.currentRepoWidget().diffArea.selectNextFile(False)

    def blameFile(self) -> None:
        self.currentRepoWidget().blameFile()

    # -------------------------------------------------------------------------
    # Help menu

    def openRescueFolder(self) -> None:
        trash = Trash.instance()
        if trash.exists():
            openFolder(str(trash.trashDir))
        else:
            showInformation(
                self,
                _("Open trash folder"),
                _("There’s no trash folder. Perhaps you haven’t discarded a change with {0} yet.", qAppName()))

    def clearRescueFolder(self) -> None:
        trash = Trash.instance()
        sizeOnDisk, patchCount = trash.size()

        if patchCount <= 0:
            showInformation(
                self,
                _("Clear trash folder"),
                _("There are no discarded changes to delete."))
            return

        humanSize = self.locale().formattedDataSize(sizeOnDisk)

        askPrompt = paragraphs(
            _n("Do you want to permanently delete <b>{n}</b> discarded patch?",
               "Do you want to permanently delete <b>{n}</b> discarded patches?", patchCount),
            _("This will free up {0} on disk.", escape(humanSize)),
            _("This cannot be undone!")
        )

        askConfirmation(
            parent=self,
            title=_("Clear trash folder"),
            text=askPrompt,
            callback=lambda: trash.clear(),
            okButtonText=_("Delete permanently"),
            okButtonIcon=stockIcon("SP_DialogDiscardButton"))

    # -------------------------------------------------------------------------
    # File menu callbacks

    def newRepo(self) -> None:
        runner = RepoTaskRunner(self)
        runner.ready.connect(runner.deleteLater)
        NewRepo.invoke(runner, self.openRepo)

    def cloneDialog(self, initialUrl: str = "") -> None:
        dlg = CloneDialog(initialUrl, self)
        dlg.cloneSuccessful.connect(lambda path: self.openRepo(path, exactMatch=True))
        dlg.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        dlg.show()

    def openDialog(self) -> None:
        qfd = PersistentFileDialog.openDirectory(self, "NewRepo", _("Open repository"))
        qfd.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)  # don't leak dialog
        qfd.fileSelected.connect(lambda path: self.openRepo(path, exactMatch=False))
        qfd.show()

    # -------------------------------------------------------------------------
    # Tab management

    def closeCurrentTab(self) -> None:
        if self.tabs.count() == 0:  # don't attempt to close if no tabs are open
            QApplication.beep()
            return

        self.closeTab(self.tabs.currentIndex())

    def closeTab(self, index: int, finalTab: bool = True) -> None:
        widget = self.tabs.widget(index)

        # Remove the tab BEFORE cleaning up the widget
        # to prevent any interaction with it while it's wrapping up.
        self.tabs.removeTab(index)
        self.refreshAllTabTexts()

        # Clean up the widget
        widget.close()  # will call RepoWidget.cleanup()
        widget.deleteLater()  # help out GC (for PySide6)
        del widget

        if finalTab:
            self.saveSession()
            gc.collect()

    def closeOtherTabs(self, index: int) -> None:
        # First, set this tab as active so an active tab that gets closed doesn't trigger other tabs to load.
        self.tabs.setCurrentIndex(index)

        # Now close all tabs in reverse order but skip the index we want to keep.
        start = self.tabs.count()-1
        final = 1 if index == 0 else 0
        for i in range(start, -1, -1):
            if i != index:
                self.closeTab(i, i == final)

    def unloadOtherTabs(self, index: int = -1) -> None:
        if index < 0:
            index = self.tabs.currentIndex()

        # First, set this tab as active so an active tab that gets closed doesn't trigger other tabs to load.
        self.tabs.setCurrentIndex(index)

        # Now unload all tabs but skip the index we want to keep.
        numUnloaded = 0
        for i in range(self.tabs.count()):
            rw = self.tabs.widget(i)
            if i == index or not isinstance(rw, RepoWidget):
                continue
            with QSignalBlockerContext(rw):
                stub = rw.replaceWithStub()
            stub.disableAutoLoad()
            self.replaceRepoWidgetWithStub(rw, stub)
            del rw
            numUnloaded += 1

        self.statusBar2.showMessage(_n("{n} background tab unloaded.", "{n} background tabs unloaded.", numUnloaded))
        gc.collect()

    def closeAllTabs(self) -> None:
        start = self.tabs.count() - 1
        with QSignalBlockerContext(self.tabs):  # Don't let awaken unloaded tabs
            for i in range(start, -1, -1):  # Close tabs in reverse order
                self.closeTab(i, i == 0)

        self.onTabCurrentWidgetChanged()

    def reloadAllTabs(self) -> None:
        self.unloadOtherTabs()
        rw = self.currentRepoWidget()
        rw.replaceWithStub()

    def refreshAllTabTexts(self) -> None:
        widgets = list(self.tabs.widgets())
        baseTitles = [widget.getTitle() for widget in widgets]
        newTitles = baseTitles[:]

        groupsByTitle: dict[str, list[int]] = {}
        for i, title in enumerate(baseTitles):
            groupsByTitle.setdefault(title, []).append(i)

        for group in groupsByTitle.values():
            if len(group) <= 1:
                continue
            # Only disambiguate default (basename) titles; custom nicknames are kept as-is.
            defaultIndices = [
                i for i in group
                if not settings.history.getRepoNickname(widgets[i].workdir, strict=True)]
            if len(defaultIndices) <= 1:
                continue
            workdirs = [widgets[i].workdir for i in defaultIndices]
            disambiguatedTitles = disambiguateTabTitlesByPath(workdirs)
            for idx, title in zip(defaultIndices, disambiguatedTitles, strict=True):
                newTitles[idx] = title

        for i, title in enumerate(newTitles):
            self.tabs.setTabText(i, escamp(title))

        # Toggle toolbar's borderless property (for custom theme)
        borderlessToolBar = self.tabs.count() >= 2 or not self.tabs.tabs.autoHide()
        toggleQssProperty(self.mainToolBar, "borderless", borderlessToolBar)

    def openRepoNextTo(self, rw, path: str, locator: NavLocator = NavLocator.Empty):
        index = self.tabs.indexOf(rw)
        if index >= 0:
            index += 1
        return self._openRepo(path, tabIndex=index, exactMatch=True, locator=locator)

    def nextTab(self) -> None:
        if self.tabs.count() == 0:
            QApplication.beep()
            return
        index = self.tabs.currentIndex()
        index += 1
        index %= self.tabs.count()
        self.tabs.setCurrentIndex(index)

    def previousTab(self) -> None:
        if self.tabs.count() == 0:
            QApplication.beep()
            return
        index = self.tabs.currentIndex()
        index += self.tabs.count() - 1
        index %= self.tabs.count()
        self.tabs.setCurrentIndex(index)

    # -------------------------------------------------------------------------
    # Session management

    def restoreSession(self, session: settings.Session, sloppyPaths: list[str] | None = None) -> None:
        # Note: window geometry, despite being part of the session file, is
        # restored in application.py to avoid flashing a window with incorrect
        # dimensions on boot

        RepoWidget.sharedSplitterSizes = copy.deepcopy(session.splitterSizes)

        # Stop here if there are no tabs to load
        if not session.tabs:
            return

        errors = []

        # We might not be able to load all tabs, so we may have to adjust the active tab index.
        activeTab = -1
        successfulRepos = []

        # Lazy-loading: prepare all tabs, but don't load the repos (foreground=False).
        for i, path in enumerate(session.tabs):
            sloppy = sloppyPaths is not None and path in sloppyPaths

            try:
                newRepoWidget = self._openRepo(path, exactMatch=not sloppy, foreground=False)
            except (GitError, OSError, NotImplementedError) as exc:
                # GitError: most errors thrown by pygit2
                # OSError: e.g. permission denied
                # NotImplementedError: e.g. shallow/bare repos
                errors.append((path, exc))
                continue

            assert isinstance(newRepoWidget, RepoWidget | RepoStub)

            # If we were passed a "sloppy" path from the command line, remember the root path.
            if sloppy:
                path = newRepoWidget.workdir

            successfulRepos.append(path)

            if i == session.activeTabIndex:
                # Heads up: MainWindow._openRepo may return an existing RepoWidget that matches the
                # given path. So, we're not necessarily the last tab, e.g. if the user passes
                # duplicate paths on the CLI.
                activeTab = self.tabs.indexOf(newRepoWidget)

        # If we failed to load anything, tell the user about it
        if errors:
            self._reportSessionErrors(errors)

        # Update history (don't write it yet - onTabCurrentWidgetChanged will do it below)
        for path in reversed(successfulRepos):
            settings.history.addRepo(path)
        self.fillRecentMenu()

        # Fall back to tab #0 if desired tab couldn't be restored (otherwise welcome page will stick around)
        if activeTab < 0 and len(successfulRepos) >= 1:
            activeTab = 0

        # Set current tab and load its repo.
        if activeTab >= 0:
            self.tabs.setCurrentIndex(activeTab)
            self.onTabCurrentWidgetChanged()  # needed to trigger loading on tab #0

    def _reportSessionErrors(self, errors: Sequence[tuple[str, BaseException]]) -> None:
        numErrors = len(errors)
        text = _n("The session couldn’t be restored fully because a repository failed to load:",
                  "The session couldn’t be restored fully because {n} repositories failed to load:", numErrors)
        qmb = asyncMessageBox(self, 'warning', _("Restore session"), text)
        addULToMessageBox(qmb, [f"<b>{compactPath(path)}</b><br>{exc}" for path, exc in errors])
        qmb.show()

    def saveSession(self, writeNow=False) -> None:
        session = settings.Session()
        session.windowGeometry = self.saveGeometry().data()
        session.splitterSizes = RepoWidget.sharedSplitterSizes.copy()
        session.tabs = [widget.workdir for widget in self.tabs.widgets()]
        session.activeTabIndex = self.tabs.currentIndex()
        session.setDirty()
        if writeNow:
            session.write()

    def closeEvent(self, event: QCloseEvent) -> None:
        if not GFApplication.instance().mountManager.checkOnClose(self, self.close):
            event.setAccepted(False)
            return

        # Save session before closing all tabs.
        self.saveSession(writeNow=True)

        # Close all tabs so RepoWidgets release all their resources.
        # Important so unit tests wind down properly!
        self.closeAllTabs()

        super().closeEvent(event)

    # -------------------------------------------------------------------------
    # Drag and drop

    def getDropOutcomeFromLocalFilePath(self, path: str) -> tuple[DropAction, str]:
        if path.endswith(".patch"):
            return DropAction.Patch, path

        try:
            workdir, existingWidget = self._resolveWorkdir(path, exactMatch=False)
        except Exception as exc:
            if isinstance(exc, GitError) and str(exc).startswith("Repository not found"):
                return DropAction.Deny, _("{0} isn’t in a Git repo", tquoe(Path(path).name))
            else:  # pragma: no cover
                logger.exception("in drop outcome: _resolveWorkdir failed")
                return DropAction.Deny, str(exc)

        if existingWidget is not None and existingWidget.isVisible() and Path(path).is_file():
            return DropAction.Blame, path

        return DropAction.Open, workdir

    def getDropOutcomeFromMimeData(self, mime: QMimeData) -> tuple[DropAction, str]:
        if mime.hasUrls():
            try:
                url: QUrl = mime.urls()[0]
            except IndexError:
                return DropAction.Deny, ""

            if url.isLocalFile():
                path = url.toLocalFile()
                return self.getDropOutcomeFromLocalFilePath(path)
            else:
                return DropAction.Clone, url.toString()

        elif mime.hasText():
            text = mime.text()
            text = text.strip()
            if os.path.isabs(text) and os.path.exists(text):
                return self.getDropOutcomeFromLocalFilePath(text)
            elif text.startswith(("ssh://", "git+ssh://", "https://", "http://")):
                return DropAction.Clone, text
            elif re.match(r"^[a-zA-Z0-9-_.]+@.+:.+", text):
                return DropAction.Clone, text
            else:
                return DropAction.Deny, ""

        return DropAction.Deny, ""

    def handleDrop(self, action: DropAction, data: str) -> None:
        if action == DropAction.Deny:
            pass
        elif action == DropAction.Clone:
            self.cloneDialog(data)
        elif action == DropAction.Blame:
            rw = self.currentRepoWidget()
            path = Path(data)
            path = path.relative_to(rw.workdir)  # May raise ValueError('X is not in the subpath of Y')
            pathStr = path.as_posix()  # WINDOWS: Convert to forward slashes to match internal git representation
            rw.blameFile(pathStr)
        elif action == DropAction.Open:
            self.openRepo(data, exactMatch=True)
        elif action == DropAction.Patch:
            tasks.ApplyPatchFile.invoke(self, data)
        else:
            warnings.warn(f"Unsupported drag-and-drop outcome {action}")  # type: ignore[unreachable]

    # -------------------------------------------------------------------------
    # Prefs

    def refreshPrefs(self) -> None:
        self.statusBar2.setVisible(settings.prefs.showStatusBar)
        self.statusBar2.enableMemoryIndicator(APP_DEBUG)
        self.mainToolBar.setVisible(settings.prefs.showToolBar)
        self.showStatusBarAction.setChecked(settings.prefs.showStatusBar)
        self.showMenuBarAction.setChecked(settings.prefs.showMenuBar)

    def onApplyPrefs(self, changedKeys: set[str]) -> None:
        if "showMenuBar" in changedKeys and not settings.prefs.showMenuBar:
            self.showMenuBarHiddenWarning()

        if PrefEffects.RebuildMenu & changedKeys:
            self.fillGlobalMenuBar()

        if PrefEffects.RestartApp & changedKeys:
            showInformation(
                self,
                _("Apply Settings"),
                _("You may need to restart {app} for the new settings to take effect fully.", app=qAppName()))
        elif PrefEffects.ReloadRepo & changedKeys and self.tabs.count() != 0:
            qmb = asyncMessageBox(
                self,
                "question",
                _("Apply Settings"),
                _("The new settings won’t take effect fully until you reload the current repositories."),
                buttons=QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
            qmb.button(QMessageBox.StandardButton.Ok).setText(_("&Reload"))
            qmb.button(QMessageBox.StandardButton.Cancel).setText(_("&Not Now"))
            qmb.accepted.connect(self.reloadAllTabs)
            qmb.show()

        # If any changed setting matches autoReload, schedule a forced refresh
        # of the current diff in all loaded RepoWidgets.
        if PrefEffects.ReloadDiff & changedKeys:
            # Nuke cached syntax highlighting
            LexJobCache.clear()

            for rw in self.tabs.widgets():
                if not isinstance(rw, RepoWidget):
                    continue
                locator = rw.taskRunner.pendingEpilog.jumpTo or rw.navLocator
                locator = locator.withExtraFlags(NavFlags.ForceDiff | NavFlags.ForceRecreateDocument)
                rw.taskRunner.pendingEpilog.jumpTo = locator
                rw.refreshRepo()

    # -------------------------------------------------------------------------
    # Dispatch commands to detached windows

    def dispatchCloseCommand(self) -> None:
        if self.isActiveWindow():
            self.closeCurrentTab()
            return

        # This is for macOS. Systems without a global main menu (i.e. anything but macOS)
        # take a different path to intercept keyboard shortcuts.
        try:
            CodeView.currentDetachedCodeView().window().close()
        except KeyError:
            QApplication.beep()

    def dispatchSearchCommand(self, op: SearchBar.Op = SearchBar.Op.Start) -> None:
        if self.isActiveWindow() and self.currentRepoWidget():
            self.currentRepoWidget().dispatchSearchCommand(op)
            return

        # This is for macOS. Systems without a global main menu (i.e. anything but macOS)
        # take a different path to intercept keyboard shortcuts.
        try:
            CodeView.currentDetachedCodeView().searchBar.popUp(op)
        except KeyError:
            QApplication.beep()

    # -------------------------------------------------------------------------
    # User commands

    def parseUserCommands(self) -> None:
        self.userCommands = list(UserCommand.parseCommandBlock(settings.prefs.commands))

    def contextualUserCommands(self, *placeholderTokens: UserCommand.Token) -> list[ActionDef]:
        tokenSet = set(placeholderTokens)
        actions: list[ActionDef] = []
        for command in self.userCommands:
            if not command.matchesContext(tokenSet):
                continue
            if not actions:
                actions.append(ActionDef.SEPARATOR)
            actions.append(ActionDef(
                _("(Command) {0}", command.menuTitle()),
                lambda c=command: self.currentRepoWidget().executeUserCommand(c),
                "prefs-usercommands",
                tip=command.menuToolTip(),
                shortcuts=command.shortcut,
            ))
        return actions

    # -------------------------------------------------------------------------
    # Repository-less actions (actions that only need a path to the workdir;
    # full-blown RepoWidget not required)

    def repolessActions(self, workdir: str | Callable[[], str]) -> list[ActionDef]:
        superprojectLabel = _("Open Superproject")
        superprojectEnabled = True
        superproject = None

        if isinstance(workdir, str):
            def workdirProxy():
                return workdir
            superproject = settings.history.getRepoSuperproject(workdir)
            superprojectEnabled = bool(superproject)
            if superprojectEnabled:
                superprojectName = settings.history.getRepoTabName(superproject)
                superprojectLabel = _("Open Superproject {0}", lquo(superprojectName))
        else:
            workdirProxy = workdir
        assert callable(workdirProxy)

        return [
            ActionDef(
                _("&Open Repo Folder"),
                lambda: openFolder(workdirProxy()),
                icon="reveal",
                shortcuts=GlobalShortcuts.openRepoFolder,
                tip=_("Open this repo’s working directory in the system’s file manager"),
            ),

            ActionDef(
                _("Open &Terminal"),
                lambda: ToolProcess.startTerminal(self, workdirProxy()),
                icon="terminal",
                shortcuts=GlobalShortcuts.openTerminal,
                tip=_("Open a terminal in the repo’s working directory"),
            ),

            ActionDef(
                _("Op&en Repo in {0}", getExternalEditorName()),
                lambda: self.openRepoInEditor(workdirProxy()),
                icon="prefs-diff",
                shortcuts=GlobalShortcuts.NO_SHORTCUT,
                tip=_("Open this repo’s working directory in {0}", getExternalEditorName()),
            ),

            ActionDef(
                _("Cop&y Repo Path"),
                lambda: self.repolessCopyPath(workdirProxy()),
                tip=_("Copy the absolute path to this repo’s working directory to the clipboard"),
            ),

            ActionDef(
                superprojectLabel,
                lambda: self.repolessOpenSuperproject(workdirProxy(), superproject),
                enabled=superprojectEnabled,
            ),

            ActionDef(
                _("Re&name\u2026"),
                lambda: self.repolessSetNickname(workdirProxy()),
                icon="rename",
            ),
        ]

    def openRepoInEditor(self, workdir: str) -> None:
        if settings.prefs.externalEditor == "":
            setUpToolCommand(self, "externalEditor")
            return

        ToolProcess.startTextEditor(self, workdir)

    def repolessSetNickname(self, workdir: str) -> None:
        defaultName = Path(workdir).name
        oldNickname = settings.history.getRepoNickname(workdir, strict=True)

        title = _("Nickname for {0}", tquoe(defaultName))
        label = _("Enter or clear nickname for this repository:")
        leaveBlank = _("Leave blank to refer to this repo as {0}", tquoe(defaultName))
        hint = "<p>" + _p(  # Sync with ui_reposettingsdialog.py to keep existing translation
            "RepoSettingsDialog",
            "This nickname will appear within {app} in tab names, menus, etc. "
            "It does not change the actual name of the repo’s directory. "
            "Leave blank to clear the nickname.",
            app=qAppName())

        tid = TextInputDialog(self, title, label, subtitle=escamp(compactPath(workdir)), hint=hint)
        tid.lineEdit.setClearButtonEnabled(True)
        tid.lineEdit.setPlaceholderText(leaveBlank)
        tid.lineEdit.setText(oldNickname)
        tid.lineEdit.selectAll()
        tid.textAccepted.connect(lambda nick: self._repolessSetNicknameImpl(workdir, nick))
        tid.show()

    def _repolessSetNicknameImpl(self, workdir: str, nick: str) -> None:
        settings.history.setRepoNickname(workdir, nick)
        settings.history.setDirty()
        tabWidget = self.tabWidgetForWorkdirPath(workdir)
        if isinstance(tabWidget, RepoWidget):
            # Percolate to sidebar, etc.
            tabWidget.nameChange.emit()
        else:
            # RepoStub
            self.onRepoNameChanged()

    def repolessCopyPath(self, workdir: str) -> None:
        QApplication.clipboard().setText(workdir)
        self.statusBar2.showMessage(clipboardStatusMessage(workdir))

    def repolessOpenSuperproject(self, workdir: str, superproject: str | None = None) -> None:
        if superproject is None:
            superproject = settings.history.getRepoSuperproject(workdir)

        if not superproject:
            showInformation(self, _("Open Superproject"), _("This repository does not have a superproject."))
            return

        submoduleTab = self.tabWidgetForWorkdirPath(workdir)  # may be None
        self.openRepoNextTo(submoduleTab, superproject)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import html
import io
import logging
import re
import shlex
import signal
from enum import StrEnum
from pathlib import Path
from typing import ClassVar

from gitfourchette import settings
from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.gitdriver.gitdelta import GitDelta, GitStatus
from gitfourchette.gitdriver.gitdeltafile import GitDeltaSource
from gitfourchette.gitdriver.lfspointer import LfsObjectCacheMissingError
from gitfourchette.gitdriver.parsers import parseGitStatus, parseGitDiffRawZ
from gitfourchette.porcelain import version_to_tuple, Oid, EMPTYTREE_OID
from gitfourchette.qt import *
from gitfourchette.settings import WhitespaceMode

logger = logging.getLogger(__name__)


def argsIf(condition: bool, *args: str) -> tuple[str, ...]:
    if condition:
        return args
    else:
        return ()


class VanillaFetchStatusFlag(StrEnum):
    FastForward = " "
    ForcedUpdate = "+"
    PrunedRef = "-"
    TagUpdate = "t"
    NewRef = "*"
    Rejected = "!"
    UpToDate = "="


class GitDriver(QProcess):
    _commandStem            : ClassVar[list[str]] = ["/usr/bin/git"]
    _cachedGitVersionValid  : ClassVar[bool] = False
    _cachedGitVersion       : ClassVar[str] = ""
    _cachedGitVersionTuple  : ClassVar[tuple[int, ...]] = (0,)
    _cachedLfsVersionValid  : ClassVar[bool] = False
    _cachedLfsVersion       : ClassVar[str] = ""

    progressMessage = Signal(str)
    progressFraction = Signal(int, int)

    @classmethod
    def runSync(
            cls,
            *args: str,
            directory: str = "",
            strict: bool = False
    ):
        return ToolCommands.runSync(*cls._commandStem, *args, directory=directory, strict=strict)

    @classmethod
    def setGitPath(cls, gitPath: str):
        cls._commandStem = ToolCommands.splitCommandTokens(gitPath)
        cls._cachedGitVersionValid = False

    @classmethod
    def _cacheGitVersion(cls, rawVersionText: str = ""):
        if cls._cachedGitVersionValid:
            return

        if not rawVersionText:
            rawVersionText = cls.runSync("version")
        text = rawVersionText.removeprefix("git version").strip()

        try:
            numberStr = text.split(maxsplit=1)[0]
        except IndexError:
            numberStr = "0"

        cls._cachedGitVersionValid = True
        cls._cachedGitVersion = text
        cls._cachedGitVersionTuple = version_to_tuple(numberStr)

    @classmethod
    def gitVersion(cls) -> str:
        cls._cacheGitVersion()
        return cls._cachedGitVersion

    @classmethod
    def gitVersionTuple(cls) -> tuple[int, ...]:
        cls._cacheGitVersion()
        return cls._cachedGitVersionTuple

    @classmethod
    def _cacheLfsVersion(cls):
        if cls._cachedLfsVersionValid:
            return

        # strict=False because it doesn't have to be installed
        rawVersionText = cls.runSync("lfs", "version", strict=False)
        text = rawVersionText.removeprefix("git-lfs/").strip()
        text = text.split(' ', 1)[0]

        cls._cachedLfsVersion = text
        cls._cachedLfsVersionValid = True

    @classmethod
    def lfsVersion(cls) -> str:
        cls._cacheLfsVersion()
        return cls._cachedLfsVersion

    @classmethod
    def supportsFetchPorcelain(cls) -> bool:
        # fetch --porcelain is only available since git 2.41 (June 2023)
        # Ubuntu 22.04 LTS ships with git 2.34.1.
        # Debian 12 and macOS 15 ship with git 2.39.5.
        return cls.gitVersionTuple() >= (2, 41)

    @classmethod
    def supportsDashDashBeforePositionalArgs(cls) -> bool:
        """
        True if git commands that ONLY take positional arguments (no switches)
        will accept a double dash before the argument list.
        Example: 'git remote remove -- <name>'
        (Note: Commands that accept switches always accept '--')
        """
        # Note: I haven't bothered bisecting the exact version of git that
        # introduced support for this. It appeared somewhere between 2.34 and
        # 2.39 (the next version above 2.34 that I easily had access to).
        return cls.gitVersionTuple() >= (2, 39)

    @classmethod
    def parseTable(cls, pattern: str, stdout: str, linesep="\n", strict=True) -> list:
        table = []

        stdout = stdout.removesuffix(linesep)

        for line in stdout.split(linesep):
            match = re.match(pattern, line)

            if match is None:
                if strict:
                    raise ValueError("table line does not match pattern: " + line)
                else:
                    continue

            table.append(match.groups())

        return table

    def __init__(self, *args: str, parent: QObject | None = None):
        super().__init__(parent)

        self.setObjectName("GitDriver")

        tokens = GitDriver._commandStem + list(args)
        self.setProgram(tokens[0])
        self.setArguments(tokens[1:])

        self.readyReadStandardError.connect(self._onReadyReadStandardError)
        self._stderrScrollback = io.BytesIO()
        self._stdout: str | None = None

    def stdoutTable(self, pattern: str, linesep="\n", strict=True) -> list:
        stdout = self.stdoutScrollback()
        return self.parseTable(pattern, stdout, linesep, strict)

    def stdoutTableNumstatZ(self, strict=True) -> list[tuple[str, str, str]]:
        pattern = r"^(-|\d+)\t(-|\d+)\t(.+)$"
        stdout = self.stdoutScrollback()
        return self.parseTable(pattern, stdout, "\0", strict)

    @classmethod
    def parseProgress(cls, stderr: bytes | str) -> tuple[str, int, int]:
        text = ""
        num = -1
        denom = -1

        if isinstance(stderr, bytes):
            stderr = stderr.decode("utf-8", errors="replace")
        lines = stderr.splitlines()

        if lines:
            # Report last line
            text = lines[-1]

            # Look for a fraction, e.g. "(50/1000)"
            for line in lines:
                fractionMatch = re.search(r"\((\d+)/(\d+)\)", line)
                if fractionMatch:
                    num = int(fractionMatch.group(1))
                    denom = int(fractionMatch.group(2))

        return text, num, denom

    @classmethod
    def reformatHintText(cls, stderr: str):
        previousTag = ""
        parts: list[str] = []

        for stderrLine in stderr.splitlines():
            try:
                tag, text = stderrLine.split(":", 1)
                tag = html.escape(tag)
                text = html.escape(text)
            except ValueError:
                tag, text = "", stderrLine

            if tag != previousTag:
                if parts:
                    parts.append("<br>")
                if tag:
                    parts.append(f"<b>{tag}:</b>")

            parts.append(text)
            previousTag = tag

        return "".join(parts)

    def _onReadyReadStandardError(self):
        raw = self.readAllStandardError().data()
        self._stderrScrollback.write(raw)

        text, num, denom = GitDriver.parseProgress(raw)
        if text:
            self.progressMessage.emit(text)
        if num >= 0 and denom >= 0:
            self.progressFraction.emit(num, denom)

    def stderrScrollback(self) -> str:
        return '\n'.join(
            line.rstrip().decode("utf-8", errors="replace")
            for line in self._stderrScrollback.getvalue().splitlines(keepends=True)
            if not line.endswith(b"\r")
        )

    def stdoutScrollback(self) -> str:
        if self._stdout is None:
            self._stdout = self.readAllStandardOutput().data().decode("utf-8", errors="replace")
        return self._stdout

    def readPostCommitInfo(self) -> tuple[str, str]:
        # [master 123abc]
        # [master (root-commit) 123abc]
        # [detached HEAD 123abc]
        stdout = self.stdoutScrollback()
        match = re.match(r"^\[(.+)\s+([\da-f]+)]", stdout, re.IGNORECASE)
        if not match:
            raise ValueError("couldn't parse post-commit stdout: " + stdout.splitlines()[0])
        branchName = match.group(1)
        commitHash = match.group(2)
        return branchName, commitHash

    def readFetchPorcelainUpdatedRefs(self) -> dict[str, tuple[str, Oid, Oid]]:
        """
        Read a table of updated refs from the output of "git fetch --porcelain".
        Requires git 2.41.
        """
        assert self.supportsFetchPorcelain(), "did you forget to gate this call with GitDriver.supportsFetchPorcelain()?"

        table = self.stdoutTable(r"^(.) ([\da-f]+) ([\da-f]+) (.+)$", strict=False)

        return {
            localRef: (flag, Oid(hex=oldHex), Oid(hex=newHex))
            for flag, oldHex, newHex, localRef in table
        }

    def readStatusPorcelainV2Z(self, sourceCommit: Oid | None) -> tuple[int, list[GitDelta], list[GitDelta]]:
        stdout = self.stdoutScrollback()
        parser = parseGitStatus(stdout, self.workingDirectory())
        stagedDeltas = []
        unstagedDeltas = []
        numEntries = 0
        for staged, unstaged in parser:
            numEntries += 1
            if staged is not None:
                staged.old.sourceCommit = sourceCommit
                stagedDeltas.append(staged)
            if unstaged is not None:
                unstagedDeltas.append(unstaged)
        return numEntries, stagedDeltas, unstagedDeltas

    @classmethod
    def buildDiffRawCommand(cls, delta: tuple[Oid | None, Oid | None]) -> list[str]:
        assert isinstance(delta, tuple)
        a, b = delta
        a = a if a is not None else EMPTYTREE_OID
        b = b if b is not None else EMPTYTREE_OID

        return [
            "-c", "core.abbrev=no",
            "diff", "--raw", "-z",
            str(a),
            str(b),
        ]

    def readDiffRawZ(self) -> list[GitDelta]:
        stdout = self.stdoutScrollback()
        deltas = list(parseGitDiffRawZ(stdout))
        return deltas

    @staticmethod
    def diffFormattingArgs() -> list[str]:
        """
        Extra `git diff` arguments for line-ending / whitespace handling.
        Only applied when generating patches for on-screen display so exports,
        revert, and trash backups stay exact.
        """
        m = settings.prefs.whitespaceMode
        if m == WhitespaceMode.Strict:
            return []
        if m == WhitespaceMode.IgnoreChange:
            return ["--ignore-space-change"]
        if m == WhitespaceMode.IgnoreAll:
            return ["--ignore-all-space"]
        if m == WhitespaceMode.IgnoreCrAtEol:
            return ["--ignore-cr-at-eol"]
        raise NotImplementedError(f"unsupported WhitespaceMode {m}")

    @classmethod
    def buildDiffCommand(
            cls,
            delta: GitDelta | tuple[Oid | None, Oid | None] | None,
            binary=True,
            forDisplay: bool = False,
    ) -> list[str]:
        tokens = [
            "-c", "core.abbrev=no",
            "-c", f"diff.context={settings.prefs.contextLines}",
            "diff",
            *argsIf(forDisplay, *cls.diffFormattingArgs()),
            *argsIf(binary, "--binary"),
        ]

        if isinstance(delta, GitDelta):
            if delta.source == GitDeltaSource.Index:
                tokens.append("--staged")
            elif delta.source == GitDeltaSource.Commit:
                # Append treeishes being compared
                assert delta.new.sourceCommit
                compareB = delta.new.sourceCommit
                compareA = delta.old.sourceCommit or EMPTYTREE_OID  # emptytree if root commit
                tokens += [str(compareA), str(compareB)]

            # Append paths
            if delta.status == GitStatus.Untracked:  # untracked, compare to nothing
                # --no-index is mandatory on Windows here
                tokens += ["--no-index", "--", "/dev/null", delta.new.path]
            elif delta.old.path != delta.new.path:
                tokens += ["--", delta.old.path, delta.new.path]
            else:
                tokens += ["--", delta.new.path]

        elif isinstance(delta, tuple):
            a, b = delta
            a = a if a is not None else EMPTYTREE_OID
            b = b if b is not None else EMPTYTREE_OID
            tokens.append(str(a))
            tokens.append(str(b))

        else:
            assert delta is None  # type: ignore[unreachable]

        return tokens

    @classmethod
    def buildDiffCommandLFS(cls, delta: GitDelta) -> list[str]:
        # Check that both LFS object paths are available
        missing = [
            p for p in (delta.old.lfs, delta.new.lfs)
            if p.objectPath != "/dev/null" and not Path(p.objectPath).exists()
        ]
        if missing:
            raise LfsObjectCacheMissingError(*missing)

        # Empty --git-dir argument: Prevent loading LFS attributes from the repo
        # to get a raw diff, not a diff of LFS pointers.
        tokens = [
            "--git-dir=",
            "-c", "core.abbrev=no",
            "-c", f"diff.context={settings.prefs.contextLines}",
            "diff",
            *cls.diffFormattingArgs(),
            "--no-index", "--",
            delta.old.lfs.objectPath,
            delta.new.lfs.objectPath,
        ]

        return tokens

    def formatExitCode(self) -> str:
        code = self.exitCode()

        if FLATPAK and code > 128 and self.program() == "flatpak-spawn":
            # flatpak-spawn may shift SIG numbers by 128
            # (for example: 143-128=15, a.k.a. SIGTERM)
            codeForName = code - 128
        else:
            codeForName = code

        if WINDOWS:
            code32 = code & 0xFFFFFFFF
            if code32 == 0xC000013A:
                return f"SIGTERM equivalent (0x{code32:08X})"
            elif code32 == 0xF291:
                return f"SIGKILL equivalent (0x{code32:08X})"
            else:
                return f"{code}"

        try:
            s = signal.Signals(codeForName)
            return f"{code} ({s.name})"
        except ValueError:
            pass

        return f"{code}"

    def formatCommandLine(self):
        return shlex.join([self.program()] + self.arguments())

    def htmlErrorText(self, subtitle: str = "", reformatHintText=False) -> str:
        from gitfourchette.localization import _
        from gitfourchette.toolbox import escape

        stderr = self.stderrScrollback().strip()
        if reformatHintText:
            stderr = self.reformatHintText(stderr)
        elif stderr:
            stderr = "<pre style='white-space: pre-wrap;'>" + escape(stderr)

        if subtitle:
            subtitle = f"<p>{subtitle}</p>"

        exitText = self.formatExitCode()
        if self.exitCode() == 0:
            exitText = f"<b><add>{exitText}</b></add>"
        else:
            exitText = f"<b><del>{exitText}</b></del>"

        return "".join([
            "<html style='white-space: pre-wrap;'>",
            settings.prefs.addDelColorsStyleTag(),
            "<p>",
            _("Git command exited with code {0}.", exitText),
            "</p>",
            subtitle,
            "<small>",
            stderr,
            "</html>"
        ])



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
from collections.abc import Set

from gitfourchette import colors
from gitfourchette import settings
from gitfourchette.graph import Graph, Frame
from gitfourchette.porcelain import Oid
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID

logger = logging.getLogger(__name__)

LANE_WIDTH = 10
LANE_THICKNESS = 2
DOT_RADIUS = 3
UC_COLOR = colors.gray
UC_STIPPLE = 12

_dummyEmptyList: list = []


def getColor(laneID):
    return colors.rainbowBright[laneID % len(colors.rainbowBright)]


def flattenLanes(frame: Frame, hiddenCommits: Set[Oid]) -> tuple[list[tuple[int, int]], int]:
    """
    Compute columns (horizontal positions) for each lane above and below this row.
    """

    if settings.prefs.flattenLanes:
        laneRemap, columnCount = frame.flattenLanes(hiddenCommits)
    else:
        # Straightforward lane positions (lane column == lane ID)
        laneRemap = []
        columnCount = max(len(frame.solvedArcs), len(frame.openArcs))
        for i in range(columnCount):
            laneRemap.append((i, i))

    return laneRemap, columnCount


def getCommitBulletColumn(
        commitLane: int,
        columnCount: int,
        lanePositionsAB: list[tuple[int, int]]
) -> tuple[int, int]:
    # Find out at which position to draw the commit's bullet point.
    myLanePosition = -1

    if commitLane < len(lanePositionsAB):
        posA, posB = lanePositionsAB[commitLane]

        # First, attempt to put bullet point in parent lane's position (below).
        if myLanePosition < 0:
            myLanePosition = posB

        # If that didn't work (commit has no parents), put bullet point in child lane's position (above).
        if myLanePosition < 0:
            myLanePosition = posA

    # If that still didn't work, we have a lone commit without parents or children; just toss the bullet to the right.
    if myLanePosition < 0:
        myLanePosition = columnCount
        columnCount += 1
        assert myLanePosition == commitLane, "expecting GraphGenerator to put lone commits on the rightmost column"

    return myLanePosition, columnCount


def paintGraphFrame(
        painter: QPainter,
        rect: QRect,
        oid: Oid,
        graph: Graph,
        hiddenCommits: Set[Oid]
):
    try:
        # Get this commit's sequential index in the graph
        myRow = graph.getCommitRow(oid)
    except LookupError:  # pragma: no cover
        logger.warning(f"Skipping unregistered commit: {oid}")
        return

    painter.save()
    outlineColor = painter.background().color()

    # Lines are drawn with SquareCap to fill in gaps at fractional display scaling factors.
    # This may cause the painter to overflow to neighboring rows, so set a clip rect.
    painter.setClipRect(rect)

    # Ensure all coordinates below are integers so our straight lines don't look blurry
    x = int(rect.left() + LANE_WIDTH // 2)
    top = int(rect.y())
    bottom = int(rect.y() + rect.height())  # Don't use rect.bottom(), which for historical reasons doesn't return what we want (see Qt docs)
    middle = (top + bottom) // 2

    # Get graph frame for this row
    frame = graph.getFrame(myRow)
    assert frame.commit == oid
    assert frame.row == myRow

    # Get the commit's lane ID
    commitLane = frame.homeLane()

    # Flatten the lanes so there are no horizontal gaps in-between the lanes (optional).
    # laneColumnsAB is a table of lanes to columns (horizontal positions).
    laneColumnsAB, numFlattenedColumns = flattenLanes(frame, hiddenCommits)

    # Get column (horizontal position) of commit bullet point.
    myColumn, numFlattenedColumns = getCommitBulletColumn(commitLane, numFlattenedColumns, laneColumnsAB)

    rect.setRight(x + (numFlattenedColumns - 1) * LANE_WIDTH)
    mx = x + myColumn * LANE_WIDTH  # the screen X of this commit's bullet point

    # draw bullet point _outline_ for this commit, beneath everything else
    painter.setPen(QPen(outlineColor, 2, Qt.PenStyle.SolidLine))
    painter.setBrush(Qt.BrushStyle.NoBrush)
    painter.drawEllipse(QPoint(mx, middle), DOT_RADIUS, DOT_RADIUS)

    path = QPainterPath()

    def submitPath(path: QPainterPath, column, stipple=False, dashOffset = 0):
        assert not path.isEmpty()
        # white outline
        painter.setPen(QPen(outlineColor, LANE_THICKNESS + 2, Qt.PenStyle.SolidLine, Qt.PenCapStyle.SquareCap, Qt.PenJoinStyle.BevelJoin))
        painter.drawPath(path)
        # actual color
        color = UC_COLOR if stipple else getColor(column)
        cap = Qt.PenCapStyle.FlatCap if stipple else Qt.PenCapStyle.SquareCap
        pen = QPen(color, LANE_THICKNESS, Qt.PenStyle.SolidLine, cap, Qt.PenJoinStyle.BevelJoin)
        if stipple:
            interval = rect.height()/(UC_STIPPLE*LANE_THICKNESS)
            pen.setDashPattern([interval, interval])
            pen.setDashOffset(dashOffset * 0.5 * rect.height()/(LANE_THICKNESS))
        painter.setPen(pen)
        painter.drawPath(path)
        # clear path for next iteration
        path.clear()

    if frame.commit in hiddenCommits:
        arcsPassingByCommit = _dummyEmptyList
        arcsOpenedByCommit = _dummyEmptyList
        arcsClosedByCommit = _dummyEmptyList
        junctionsAtCommit = _dummyEmptyList
    else:
        arcsPassingByCommit = frame.arcsPassingByCommit(hiddenCommits)
        arcsOpenedByCommit = list(frame.arcsOpenedByCommit(hiddenCommits))
        arcsClosedByCommit = list(frame.arcsClosedByCommit(hiddenCommits))
        junctionsAtCommit = frame.junctionsAtCommit(hiddenCommits)

    # draw arcs PASSING BY commit
    cy1 = middle
    cy2 = middle
    if not arcsOpenedByCommit:
        # Compress the curvature of arcs passing by a root commit    |  |  |
        # to the bottom half of the row to prevent these arcs from   O  |  |
        # looking like they are joined to the root commit in a         /  /
        # flattened graph.                                            |  |
        cy1 = bottom + 4
        cy2 = middle
    elif not arcsClosedByCommit:
        # Same, but compress above
        cy1 = middle - 4
        cy2 = top
    for arc in arcsPassingByCommit:
        columnA, columnB = laneColumnsAB[arc.lane]  # column above, column below
        ax = x + columnA * LANE_WIDTH
        bx = x + columnB * LANE_WIDTH
        path.moveTo(ax, top)
        path.cubicTo(ax, cy1, bx, cy2, bx, bottom)
        submitPath(path, arc.lane, arc.openedBy == UC_FAKEID)

    # draw arcs CLOSED BY commit (from above)
    for arc in reversed(arcsClosedByCommit):
        columnA, _dummy = laneColumnsAB[arc.lane]  # column above, column below
        ax = x + columnA * LANE_WIDTH
        # Path from above does elbow shape to merge into commit bullet point
        path.moveTo(ax, top)
        path.quadTo(ax, middle, mx, middle)
        submitPath(path, arc.lane, arc.openedBy == UC_FAKEID)

    # draw arcs OPENED BY commit (downwards)
    for arc in reversed(arcsOpenedByCommit):
        _dummy, columnB = laneColumnsAB[arc.lane]  # column above, column below
        bx = x + columnB * LANE_WIDTH
        # Path forks downward from commit bullet point
        path.moveTo(mx, middle)
        path.quadTo(bx, middle, bx, bottom)
        submitPath(path, arc.lane, arc.openedBy == UC_FAKEID, dashOffset=1)

    # draw arc junctions
    for arc, _junction in junctionsAtCommit:
        columnA, columnB = laneColumnsAB[arc.lane]
        ax = x + columnA * LANE_WIDTH
        bx = x + columnB * LANE_WIDTH
        path.moveTo(mx, middle)
        path.quadTo(bx, middle, bx, bottom)
        submitPath(path, arc.lane)

    # draw bullet point for this commit
    painter.setPen(Qt.PenStyle.NoPen)
    painter.setBrush(getColor(commitLane) if oid != UC_FAKEID else UC_COLOR)
    painter.drawEllipse(QPoint(mx, middle), DOT_RADIUS, DOT_RADIUS)

    # we're done, clean up
    painter.restore()

    # add some padding to the right
    rect.setRight(rect.right() + LANE_WIDTH)




# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import enum
import filecmp
import logging
import shutil
from typing import ClassVar

from gitfourchette import settings
from gitfourchette.gitdriver import GitConflict
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *
from gitfourchette.exttools.toolprocess import ToolProcess

logger = logging.getLogger(__name__)


class MergeDriver(QObject):
    class State(enum.IntEnum):
        Idle = 0
        Busy = 1
        Fail = 2
        Tentative = 3
        Ready = 4

    @dataclasses.dataclass
    class MergeFiles:
        ours: str
        theirs: str
        ancestor: str
        scratch: str
        target: str

    _ongoingMerges: ClassVar[list[MergeDriver]] = []
    _mergeCounter: ClassVar[int] = 0

    statusChange = Signal()

    conflict: GitConflict
    paths: MergeFiles
    process: QProcess | None
    processName: str
    state: State
    debrief: str

    def __init__(self, parent: QObject, conflict: GitConflict, paths: MergeFiles):
        super().__init__(parent)

        logger.info(f"Initialize MergeDriver: {conflict}")
        self.conflict = conflict
        self.paths = paths
        self.process = None
        self.processName = "?"
        self.state = MergeDriver.State.Idle
        self.debrief = ""

        assert conflict.ours, "MergeDriver requires an 'ours' side"
        assert conflict.theirs, "MergeDriver requires a 'theirs' side"

        # Keep track of this merge
        MergeDriver._mergeCounter += 1
        MergeDriver._ongoingMerges.append(self)
        self.destroyed.connect(lambda: MergeDriver._forget(self))

    def deleteNow(self):
        MergeDriver._forget(self)
        # TODO: Terminate process?
        self.deleteLater()

    def startProcess(self, reopenWorkInProgress=False):
        paths = self.paths
        tokens = {
            "$B": paths.scratch if reopenWorkInProgress else paths.ancestor,
            "$L": paths.ours,
            "$R": paths.theirs,
            "$M": paths.scratch,
        }
        parentWidget = findParentWidget(self)
        self.process = ToolProcess.startProcess(parentWidget, ToolProcess.PrefKeyMergeTool, replacements=tokens, positional=[])
        if not self.process:
            return
        self.processName = settings.getMergeToolName()
        self.process.errorOccurred.connect(self.onMergeProcessError)
        self.process.finished.connect(self.onMergeProcessFinished)
        self.state = MergeDriver.State.Busy
        self.debrief = ""

    def onMergeProcessError(self, error: QProcess.ProcessError):
        logger.warning(f"Merge tool error {error}")

        self.state = MergeDriver.State.Fail

        if error == QProcess.ProcessError.FailedToStart:
            self.debrief = _("{0} failed to start.", tquo(self.processName))
        else:
            errorName = str(error) if PYQT5 else error.name
            self.debrief = _("{0} ran into error {1}.", tquo(self.processName), errorName)

        self.flush()

    def onMergeProcessFinished(self, exitCode: int, exitStatus: QProcess.ExitStatus):
        if exitCode != 0 or exitStatus == QProcess.ExitStatus.CrashExit:
            informalPid = self.process.processId() if self.process else '???'
            logger.warning(f"Merge tool PID {informalPid} finished with code {exitCode}, {exitStatus}")
            self.state = MergeDriver.State.Fail
            self.debrief = _("{0} didn’t complete the merge.", tquo(self.processName))
            self.debrief += "\n" + _("Exit code: {0}.", exitCode)
        else:
            self.state = MergeDriver.State.Tentative
            self.debrief = ""

        self.flush()

    def flush(self):
        if self.process is not None:
            self.process.deleteLater()
            self.process = None
        self.statusChange.emit()

    def checkUnchanged(self):
        return filecmp.cmp(self.paths.scratch, self.paths.target)

    def copyScratchToTarget(self):
        shutil.copyfile(self.paths.scratch, self.paths.target)

    @classmethod
    def findOngoingMerge(cls, conflict: GitConflict) -> MergeDriver | None:
        return next((m for m in cls._ongoingMerges if m.conflict == conflict), None)

    @classmethod
    def _forget(cls, deadMerge: MergeDriver):
        cls._ongoingMerges = [x for x in cls._ongoingMerges if x is not deadMerge]
